@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2011, 2015, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2011, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2.0,
@@ -25,6 +25,8 @@
 #include "ha_ndbcluster_glue.h"
 #include "ha_ndbcluster.h"
 #include "ndb_table_guard.h"
+#include "mysql/service_thd_alloc.h"
+#include "ndb_tdc.h"
 
 #define ERR_RETURN(err)                  \
 {                                        \
@@ -59,6 +61,14 @@ struct Ndb_fk_data : Sql_alloc
   List<Ndb_fk_item> list;
   uint cnt_child;
   uint cnt_parent;
+
+  /*
+    List to store the name of the child tables that would be updated
+    when this handler's table gets updated.
+   */
+  List<char> cascading_dependents;
+  int build_cascading_dependents_list(Ndb*, const char*, const char*,
+                                      MEM_ROOT*);
 };
 
 // Forward decl
@@ -230,50 +240,23 @@ private:
   char save_db[FN_REFLEN + 1];
 };
 
-/**
- * ndbapi want's c-strings (null terminated)
- * mysql frequently uses LEX-string...(ptr + len)
- *
- * also...they have changed between 5.1 and 5.5...
- * add a small compability-kit
- */
-static inline
+
+template <size_t buf_size>
 const char *
-lex2str(const LEX_STRING& str, char buf[], size_t len)
+lex2str(const LEX_STRING& str, char (&buf)[buf_size])
 {
-  my_snprintf(buf, len, "%.*s", (int)str.length, str.str);
+  my_snprintf(buf, buf_size, "%.*s", (int)str.length, str.str);
   return buf;
 }
 
-static inline
-const char *
-lex2str(const char * str, char buf[], size_t len)
-{
-  return str;
-}
 
-static inline
-bool
-isnull(const LEX_STRING& str)
-{
-  return str.str == 0 || str.length == 0;
-}
-
-static inline
-bool
-isnull(const char * str)
-{
-  return str == 0;
-}
-
-// copied from unused table_case_convert() in mysqld.h
 static void
 ndb_fk_casedn(char *name)
 {
-  DBUG_ASSERT(name != 0);
+  assert(name != 0);
   uint length = (uint)strlen(name);
-  DBUG_ASSERT(files_charset_info != 0 &&
-              files_charset_info->casedn_multiply == 1);
+  assert(files_charset_info != 0 &&
+         files_charset_info->casedn_multiply == 1);
   files_charset_info->cset->casedn(files_charset_info,
                                    name, length, name, length);
 }
@@ -294,6 +277,126 @@ ndb_fk_casecmp(const char* name1, const char* name2)
   return strcmp(tmp1, tmp2);
 }
 
+
+/**
+  @brief Build list of all dependent child tables that will be updated
+         during a cascading update/delete of the given table.
+
+  @param  ndb                    The Ndb Object
+  @param  dbname                 Database Name of the table.
+  @param  tabname                Name of the table
+  @param  fk_mem_root            Mem_root to store the fk data.
+  @return 0 on success
+  @return error_code on failure
+ */
+int
+Ndb_fk_data::build_cascading_dependents_list(Ndb* ndb,
+                                             const char* dbname,
+                                             const char* tabname,
+                                             MEM_ROOT* fk_mem_root)
+{
+  DBUG_ENTER("Ndb_fk_data::build_cascading_dependents_list");
+
+  /* Temporary dependent table Vector */
+  List<char> tmp_dependent_tables;
+
+  /* generate main parent full name */
+  char main_parent[FN_LEN + 1];
+  sprintf(main_parent, "%s/%s/%s", dbname, ndb->getSchemaName(), tabname);
+
+  /* Save db */
+  Ndb_db_guard db_guard(ndb);
+  NDBDICT *dict= ndb->getDictionary();
+
+  /* start looping */
+  char* dependent_table = main_parent;
+  do
+  {
+    char db_name[FN_LEN + 1];
+    const char* tab_name = fk_split_name(db_name, dependent_table);
+
+    DBUG_PRINT("debug", ("Checking for child FKs to %s.%s",
+                         db_name, tab_name));
+    setDbName(ndb, db_name);
+    Ndb_table_guard child_tab(dict, tab_name);
+    const NDBTAB* table = child_tab.get_table();
+    if (!table)
+    {
+      ERR_RETURN(dict->getNdbError());
+    }
+
+    NDBDICT::List obj_list;
+    if (dict->listDependentObjects(obj_list, *table) != 0)
+    {
+      ERR_RETURN(dict->getNdbError());
+    }
+
+    for (unsigned i = 0; i < obj_list.count; i++)
+    {
+      DBUG_PRINT("debug", ("DependentObject %d : %s, Type : %d", i,
+          obj_list.elements[i].name,
+          obj_list.elements[i].type));
+      if (obj_list.elements[i].type != NdbDictionary::Object::ForeignKey)
+        continue;
+
+      /* FK found - check its child and add it to the list */
+      NDBFK fk;
+      if (dict->getForeignKey(fk, obj_list.elements[i].name) != 0)
+      {
+        ERR_RETURN(dict->getNdbError());
+      }
+
+      const char* child_table = fk.getChildTable();
+
+      uint deleteAction = fk.getOnDeleteAction();
+      uint updateAction = fk.getOnUpdateAction();
+
+      if (ndb_fk_casecmp(child_table, dependent_table) != 0 &&
+          ndb_fk_casecmp(child_table, main_parent) != 0 &&
+          (deleteAction == NDB_FK_CASCADE  ||
+           deleteAction == NDB_FK_SET_NULL ||
+           updateAction == NDB_FK_CASCADE  ||
+           updateAction == NDB_FK_SET_NULL))
+      {
+        /*
+         The fk reference is cascade enabled and
+         the child is neither the current parent nor the main parent.
+         Add this child to the dependent list if not present already.
+        */
+        const char* tabname;
+        List_iterator_fast<char> it(cascading_dependents);
+        while ((tabname= it++))
+        {
+          if (ndb_fk_casecmp(child_table, tabname) == 0)
+          {
+            /* child table already in the list */
+            break;
+          }
+        }
+        if (tabname == NULL)
+        {
+          /* Child table not in list - add it! */
+          char* new_dependent_table = strdup_root(fk_mem_root, child_table);
+
+          tmp_dependent_tables.push_back(new_dependent_table);
+          cascading_dependents.push_back(new_dependent_table, fk_mem_root);
+
+          DBUG_PRINT("info", ("Added `%s` to dependents list", child_table));
+        }
+      }
+    }
+  } while ((dependent_table = tmp_dependent_tables.pop()));
+
+  DBUG_PRINT("exit", ("Added %d tables to the cascading tables list",
+                      cascading_dependents.elements));
+
+  /* restore db */
+  db_guard.restore();
+
+  DBUG_RETURN(0);
+}
+
+
 extern bool ndb_show_foreign_key_mock_tables(THD* thd);
 
 class Fk_util
@@ -312,7 +415,7 @@ class Fk_util
     // Push as warning if user has turned on ndb_show_foreign_key_mock_tables
     if (ndb_show_foreign_key_mock_tables(m_thd))
     {
-      push_warning(m_thd, Sql_condition::WARN_LEVEL_WARN, ER_YES, msg);
+      push_warning(m_thd, Sql_condition::SL_WARNING, ER_YES, msg);
     }
 
     // Print info to log
@@ -328,7 +431,7 @@ class Fk_util
     va_start(args,fmt);
     my_vsnprintf(msg, sizeof(msg), fmt, args);
     va_end(args);
-    push_warning(m_thd, Sql_condition::WARN_LEVEL_WARN, ER_CANNOT_ADD_FOREIGN, msg);
+    push_warning(m_thd, Sql_condition::SL_WARNING, ER_CANNOT_ADD_FOREIGN, msg);
 
     // Print warning to log
     sql_print_warning("NDB FK: %s", msg);
@@ -343,7 +446,7 @@ class Fk_util
     va_start(args,fmt);
     my_vsnprintf(msg, sizeof(msg), fmt, args);
     va_end(args);
-    push_warning(m_thd, Sql_condition::WARN_LEVEL_WARN,
+    push_warning(m_thd, Sql_condition::SL_WARNING,
                  ER_CANNOT_ADD_FOREIGN, msg);
 
     char ndb_msg[MYSQL_ERRMSG_SIZE] = {0};
@@ -353,7 +456,7 @@ class Fk_util
       const NdbError& error = dict->getNdbError();
       my_snprintf(ndb_msg, sizeof(ndb_msg),
                   "%d '%s'", error.code, error.message);
-      push_warning_printf(m_thd, Sql_condition::WARN_LEVEL_WARN,
+      push_warning_printf(m_thd, Sql_condition::SL_WARNING,
                           ER_CANNOT_ADD_FOREIGN, "Ndb error: %s", ndb_msg);
     }
     // Print error to log
@@ -483,14 +586,14 @@ class Fk_util
   {
     DBUG_ENTER("resolve_mock");
     DBUG_PRINT("enter", ("mock_name '%s'", mock_name));
-    DBUG_ASSERT(is_mock_name(mock_name));
+    assert(is_mock_name(mock_name));
 
     // Load up the mock table
     Ndb_table_guard mock_tab(dict, mock_name);
     if (!mock_tab.get_table())
     {
       error(dict, "Failed to load the listed mock table '%s'", mock_name);
-      DBUG_ASSERT(false);
+      assert(false);
       DBUG_VOID_RETURN;
     }
 
@@ -592,7 +695,7 @@ class Fk_util
       if (dict->getForeignKey(fk, element.name) != 0)
       {
         // Could not find the listed fk
-        DBUG_ASSERT(false);
+        assert(false);
         continue;
       }
 
@@ -669,7 +772,7 @@ class Fk_util
                   col_names, col_types))
       {
         error(dict, "Failed to create mock parent table '%s", mock_name);
-        DBUG_ASSERT(false);
+        assert(false);
         DBUG_RETURN(false);
       }
 
@@ -751,13 +854,12 @@ public:
                          child_id, fk_index, parent_name));
     const size_t len = my_snprintf(buf, buf_size, "NDB$FKM_%d_%u_%s",
                                    child_id, fk_index, parent_name);
-    DBUG_PRINT("info", ("len: %lu, buf_size: %lu", len, buf_size));
     if (len >= buf_size - 1)
     {
       DBUG_PRINT("info", ("Size of buffer too small"));
       DBUG_RETURN(NULL);
     }
-    DBUG_PRINT("exit", ("buf: '%s', len: %lu", buf, len));
+    DBUG_PRINT("exit", ("buf: '%s'", buf));
     DBUG_RETURN(buf);
   }
 
@@ -775,7 +877,7 @@ public:
       while ((key= it1++))
       {
         char col_name_buf[FN_REFLEN];
-        const char* col_name = lex2str(key->field_name, col_name_buf, sizeof(col_name_buf));
+        const char* col_name = lex2str(key->field_name, col_name_buf);
         col_names[i++] = strdup(col_name);
       }
       col_names[i] = 0;
@@ -801,7 +903,7 @@ public:
 
     DBUG_ENTER("mock_table::create");
     DBUG_PRINT("enter", ("mock_name: %s", mock_name));
-    DBUG_ASSERT(is_mock_name(mock_name));
+    assert(is_mock_name(mock_name));
 
     if (mock_tab.setName(mock_name))
     {
@@ -818,7 +920,7 @@ public:
       DBUG_PRINT("info", ("name: %s", col_name));
       if (mock_col.setName(col_name))
       {
-        DBUG_ASSERT(false);
+        assert(false);
         DBUG_RETURN(false);
       }
 
@@ -826,7 +928,7 @@ public:
       if (!col)
       {
         // Internal error, the two lists should be same size
-        DBUG_ASSERT(col);
+        assert(col);
         DBUG_RETURN(false);
       }
 
@@ -881,7 +983,7 @@ public:
       if (dict->getForeignKey(fk, element.name) != 0)
       {
         // Could not find the listed fk
-        DBUG_ASSERT(false);
+        assert(false);
         continue;
       }
 
@@ -914,7 +1016,7 @@ public:
       {
        // Could not open the mock table
        DBUG_PRINT("error", ("Could not open the listed mock table, ignore it"));
-       DBUG_ASSERT(false);
+       assert(false);
        continue;
       }
 
@@ -922,7 +1024,7 @@ public:
       {
         DBUG_PRINT("error", ("Failed to drop the mock table '%s'",
                               mocktab_g.get_table()->getName()));
-        DBUG_ASSERT(false);
+        assert(false);
         continue;
       }
       info("Dropped mock table '%s' - referencing table dropped", table_name);
@@ -990,7 +1092,7 @@ public:
     if (dict->getForeignKey(fk, fk_name) != 0)
     {
       error(dict, "Could not find fk '%s'", fk_name);
-      DBUG_ASSERT(false);
+      assert(false);
       DBUG_RETURN(false);
     }
 
@@ -1009,7 +1111,7 @@ public:
         if (dict->dropTableGlobal(*mocktab_g.get_table(), drop_flags) != 0)
         {
           error(dict, "Failed to drop fk mock table '%s'", parent_name);
-          DBUG_ASSERT(false);
+          assert(false);
           DBUG_RETURN(false);
         }
         // table and fk dropped
@@ -1019,7 +1121,7 @@ public:
       {
         warn("Could not open the fk mock table '%s', ignoring it...",
              parent_name);
-        DBUG_ASSERT(false);
+        assert(false);
         // fallthrough and try to drop only the fk,
       }
     }
@@ -1027,7 +1129,6 @@ public:
     if (dict->dropForeignKey(fk) != 0)
     {
       error(dict, "Failed to drop fk '%s'", fk_name);
-      DBUG_ASSERT(false);
       DBUG_RETURN(false);
     }
     DBUG_RETURN(true);
@@ -1050,7 +1151,7 @@ public:
     NdbDictionary::Dictionary::List table_list;
     if (dict->listObjects(table_list, NdbDictionary::Object::UserTable, true) != 0)
     {
-      DBUG_ASSERT(false);
+      assert(false);
       DBUG_VOID_RETURN;
     }
 
@@ -1058,7 +1159,7 @@ public:
     {
       const NdbDictionary::Dictionary::List::Element& el = table_list.elements[i];
 
-      DBUG_ASSERT(el.type == NdbDictionary::Object::UserTable);
+      assert(el.type == NdbDictionary::Object::UserTable);
 
       // Check if table is in same database as the potential new parent
       if (strcmp(new_parent_db, el.database) != 0)
@@ -1110,7 +1211,7 @@ public:
       if (dict->getForeignKey(fk, element.name) != 0)
       {
         error(dict, "Could not find the listed fk '%s'", element.name);
-        DBUG_ASSERT(false);
+        assert(false);
         continue;
       }
 
@@ -1184,6 +1285,36 @@ bool ndb_fk_util_truncate_allowed(THD* thd, NdbDictionary::Dictionary* dict,
 }
 
 
+/**
+  @brief Flush the parent table after a successful addition/deletion
+         to the Foreign Key. This is done to force reload the Parent
+         table's metadata.
+
+  @param thd            thread handle
+  @param parent_db      Parent table's database name
+  @param parent_name    Parent table's name
+  @return Void
+*/
+static void
+flush_parent_table_for_fk(THD* thd,
+                          const char* parent_db, const char* parent_name)
+{
+  DBUG_ENTER("ha_ndbcluster::flush_parent_table_for_fk");
+
+  if(Fk_util::is_mock_name(parent_name))
+  {
+    /* Parent table is mock - no need to flush */
+    DBUG_PRINT("debug", ("Parent table is a mock - skipped flushing"));
+    DBUG_VOID_RETURN;
+  }
+
+  DBUG_PRINT("debug", ("Flushing table : `%s`.`%s` ",
+                       parent_db, parent_name));
+  ndb_tdc_close_cached_table(thd, parent_db, parent_name);
+  DBUG_VOID_RETURN;
+}
+
+
 int
 ha_ndbcluster::create_fks(THD *thd, Ndb *ndb)
 {
@@ -1198,7 +1329,7 @@ ha_ndbcluster::create_fks(THD *thd, Ndb *ndb)
   List_iterator<Key> key_iterator(thd->lex->alter_info.key_list);
   while ((key=key_iterator++))
   {
-    if (key->type != Key::FOREIGN_KEY)
+    if (key->type != KEYTYPE_FOREIGN)
       continue;
 
     NDBDICT *dict= ndb->getDictionary();
@@ -1237,10 +1368,10 @@ ha_ndbcluster::create_fks(THD *thd, Ndb *ndb)
       while ((col= it1++))
       {
         const NDBCOL * ndbcol= tab->getColumn(lex2str(col->field_name,
-                                                      tmpbuf, sizeof(tmpbuf)));
+                                                      tmpbuf));
         if (ndbcol == 0)
         {
-          push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
+          push_warning_printf(thd, Sql_condition::SL_WARNING,
                               ER_CANNOT_ADD_FOREIGN,
                               "Child table %s has no column %s in NDB",
                               child_tab.get_table()->getName(), tmpbuf);
@@ -1259,7 +1390,7 @@ ha_ndbcluster::create_fks(THD *thd, Ndb *ndb)
 
     if (!child_primary_key && child_index == 0)
     {
-      push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
+      push_warning_printf(thd, Sql_condition::SL_WARNING,
                           ER_CANNOT_ADD_FOREIGN,
                           "Child table %s foreign key columns match no index in NDB",
                           child_tab.get_table()->getName());
@@ -1282,7 +1413,9 @@ ha_ndbcluster::create_fks(THD *thd, Ndb *ndb)
     }
     else
     {
-      parent_db[0]= 0;
+      /* parent db missing - so the db is same as child's */
+      my_snprintf(parent_db, sizeof(parent_db), "%*s",
+                  (int)sizeof(m_dbname), m_dbname);
     }
     if (fk->ref_table.str != 0 && fk->ref_table.length != 0)
     {
@@ -1306,7 +1439,7 @@ ha_ndbcluster::create_fks(THD *thd, Ndb *ndb)
        if (!thd_test_options(thd, OPTION_NO_FOREIGN_KEY_CHECKS))
        {
          const NdbError &error= dict->getNdbError();
-         push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
+         push_warning_printf(thd, Sql_condition::SL_WARNING,
                              ER_CANNOT_ADD_FOREIGN,
                              "Parent table %s not found in NDB: %d: %s",
                              parent_name,
@@ -1331,7 +1464,7 @@ ha_ndbcluster::create_fks(THD *thd, Ndb *ndb)
                                 child_tab.get_table()->getObjectId(),
                                 existing, parent_name))
        {
-         push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
+         push_warning_printf(thd, Sql_condition::SL_WARNING,
                              ER_CANNOT_ADD_FOREIGN,
                              "Failed to create mock parent table, too long mock name");
          DBUG_RETURN(err_default);
@@ -1340,7 +1473,7 @@ ha_ndbcluster::create_fks(THD *thd, Ndb *ndb)
                            fk->ref_columns, childcols))
        {
          const NdbError &error= dict->getNdbError();
-         push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
+         push_warning_printf(thd, Sql_condition::SL_WARNING,
                              ER_CANNOT_ADD_FOREIGN,
                              "Failed to create mock parent table in NDB: %d: %s",
                              error.code, error.message);
@@ -1351,12 +1484,12 @@ ha_ndbcluster::create_fks(THD *thd, Ndb *ndb)
        parent_tab.invalidate(); // invalidate mock table when releasing
        if (parent_tab.get_table() == 0)
        {
-         push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
+         push_warning_printf(thd, Sql_condition::SL_WARNING,
                              ER_CANNOT_ADD_FOREIGN,
                              "INTERNAL ERROR: Could not find created mock table '%s'",
                              mock_name);
          // Internal error, should be able to load the just created mock table
-         DBUG_ASSERT(parent_tab.get_table());
+         assert(parent_tab.get_table());
          DBUG_RETURN(err_default);
        }       
     }
@@ -1370,10 +1503,10 @@ ha_ndbcluster::create_fks(THD *thd, Ndb *ndb)
       while ((col= it1++))
       {
         const NDBCOL * ndbcol= tab->getColumn(lex2str(col->field_name,
-                                                      tmpbuf, sizeof(tmpbuf)));
+                                                      tmpbuf));
         if (ndbcol == 0)
         {
-          push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
+          push_warning_printf(thd, Sql_condition::SL_WARNING,
                               ER_CANNOT_ADD_FOREIGN,
                               "Parent table %s has no column %s in NDB",
                               parent_tab.get_table()->getName(), tmpbuf);
@@ -1409,7 +1542,7 @@ ha_ndbcluster::create_fks(THD *thd, Ndb *ndb)
       {
         if (parentcols[i]->isBindable(* childcols[i]) == -1)
         {
-          push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
+          push_warning_printf(thd, Sql_condition::SL_WARNING,
                               ER_CANNOT_ADD_FOREIGN,
                               "Parent column %s.%s is incompatible with child column %s.%s in NDB",
                               parent_tab.get_table()->getName(),
@@ -1423,13 +1556,14 @@ ha_ndbcluster::create_fks(THD *thd, Ndb *ndb)
 
     NdbDictionary::ForeignKey ndbfk;
     char fk_name[FN_REFLEN];
-    if (!isnull(fk->name))
+    if (fk->name.str && fk->name.length)
     {
-      my_snprintf(fk_name, sizeof(fk_name), "%s",
-                  lex2str(fk->name, tmpbuf, sizeof(tmpbuf)));
+      // The fk has a name, use it
+      lex2str(fk->name, fk_name);
     }
     else
     {
+      // The fk has no name, generate a name
       my_snprintf(fk_name, sizeof(fk_name), "FK_%u_%u",
                   parent_index ?
                   parent_index->getObjectId() :
@@ -1444,21 +1578,21 @@ ha_ndbcluster::create_fks(THD *thd, Ndb *ndb)
     ndbfk.setParent(* parent_tab.get_table(), parent_index, parentcols);
     ndbfk.setChild(* child_tab.get_table(), child_index, childcols);
 
-    switch((Foreign_key::fk_option)fk->delete_opt){
-    case Foreign_key::FK_OPTION_UNDEF:
-    case Foreign_key::FK_OPTION_NO_ACTION:
+    switch((fk_option)fk->delete_opt){
+    case FK_OPTION_UNDEF:
+    case FK_OPTION_NO_ACTION:
       ndbfk.setOnDeleteAction(NdbDictionary::ForeignKey::NoAction);
       break;
-    case Foreign_key::FK_OPTION_RESTRICT:
+    case FK_OPTION_RESTRICT:
       ndbfk.setOnDeleteAction(NdbDictionary::ForeignKey::Restrict);
       break;
-    case Foreign_key::FK_OPTION_CASCADE:
+    case FK_OPTION_CASCADE:
       ndbfk.setOnDeleteAction(NdbDictionary::ForeignKey::Cascade);
       break;
-    case Foreign_key::FK_OPTION_SET_NULL:
+    case FK_OPTION_SET_NULL:
       ndbfk.setOnDeleteAction(NdbDictionary::ForeignKey::SetNull);
       break;
-    case Foreign_key::FK_OPTION_DEFAULT:
+    case FK_OPTION_DEFAULT:
       ndbfk.setOnDeleteAction(NdbDictionary::ForeignKey::SetDefault);
       break;
     default:
@@ -1466,21 +1600,21 @@ ha_ndbcluster::create_fks(THD *thd, Ndb *ndb)
       ndbfk.setOnDeleteAction(NdbDictionary::ForeignKey::NoAction);
     }
 
-    switch((Foreign_key::fk_option)fk->update_opt){
-    case Foreign_key::FK_OPTION_UNDEF:
-    case Foreign_key::FK_OPTION_NO_ACTION:
+    switch((fk_option)fk->update_opt){
+    case FK_OPTION_UNDEF:
+    case FK_OPTION_NO_ACTION:
       ndbfk.setOnUpdateAction(NdbDictionary::ForeignKey::NoAction);
       break;
-    case Foreign_key::FK_OPTION_RESTRICT:
+    case FK_OPTION_RESTRICT:
       ndbfk.setOnUpdateAction(NdbDictionary::ForeignKey::Restrict);
       break;
-    case Foreign_key::FK_OPTION_CASCADE:
+    case FK_OPTION_CASCADE:
       ndbfk.setOnUpdateAction(NdbDictionary::ForeignKey::Cascade);
       break;
-    case Foreign_key::FK_OPTION_SET_NULL:
+    case FK_OPTION_SET_NULL:
       ndbfk.setOnUpdateAction(NdbDictionary::ForeignKey::SetNull);
       break;
-    case Foreign_key::FK_OPTION_DEFAULT:
+    case FK_OPTION_DEFAULT:
       ndbfk.setOnUpdateAction(NdbDictionary::ForeignKey::SetDefault);
       break;
     default:
@@ -1508,7 +1642,26 @@ ha_ndbcluster::create_fks(THD *thd, Ndb *ndb)
 
     if (err)
     {
-      ERR_RETURN(dict->getNdbError());
+      const NdbError err = dict->getNdbError();
+      if (err.code == 721)
+      {
+        /* An FK constraint with same name exists */
+        my_error(ER_FK_DUP_NAME, MYF(0), ndbfk.getName());
+        DBUG_RETURN(err_default);
+      }
+      else
+      {
+        /* Return the error returned by dict */
+        ERR_RETURN(err);
+      }
+    }
+
+    /* Flush the parent table out if parent is different from child */
+    if (parent_tab.get_table()->getObjectId() !=
+          child_tab.get_table()->getObjectId())
+    {
+      /* flush parent table */
+      flush_parent_table_for_fk(thd, parent_db, parent_name);
     }
   }
 
@@ -1535,7 +1688,7 @@ ha_ndbcluster::referenced_by_foreign_key()
   Ndb_fk_data *data= m_fk_data;
   if (data == 0)
   {
-    DBUG_ASSERT(false);
+    assert(false);
     DBUG_RETURN(0);
   }
 
@@ -1552,7 +1705,7 @@ ha_ndbcluster::is_child_or_parent_of_fk()
   Ndb_fk_data *data= m_fk_data;
   if (data == 0)
   {
-    DBUG_ASSERT(false);
+    assert(false);
     DBUG_RETURN(0);
   }
   
@@ -1566,58 +1719,6 @@ ha_ndbcluster::can_switch_engines()
 {
   DBUG_ENTER("ha_ndbcluster::can_switch_engines");
 
-  if (!m_table)
-  {
-    /* Require the function to be called only with open table */
-    DBUG_ASSERT(m_table);
-    DBUG_RETURN(0);
-  }
-
-  THD* thd= table->in_use;
-  if (thd == 0)
-  {
-    thd= current_thd;
-  }
-
-  if (thd == 0)
-  {
-    DBUG_RETURN(0);
-  }
-
-  DBUG_ASSERT(thd_sql_command(thd) == SQLCOM_ALTER_TABLE);
-
-  HA_CREATE_INFO &create_info= thd->lex->create_info;
-  /* Check if engine clause is present */
-  if(!create_info.db_type)
-  {
-    /*
-     If not present, allow switching engines as the alter is an
-     ALTER PARTITION command and NDB allows switching engines for
-     partition regardless of FK presence.
-
-     The NULL check is done to create_info.db_type because,
-     mysql_alter_table assigns db_type value to a duplicate local copy.
-     So, create_info.db_type here will still be null, if there was
-     no 'engine' clause in alter command.(See BUG#20180331)
-     */
-    DBUG_RETURN(1);
-  }
-
-  if (create_info.db_type->db_type == DB_TYPE_NDBCLUSTER)
-  {
-    /*
-     The Alter command has both PARTIITION and ENGINE clauses
-     but the value in ENGINE clause is NDB. Hence the ALTER is
-     effectively a PARTITION command.
-     So allow switching, regardless of any FK presence
-    */
-    DBUG_RETURN(1);
-  }
-
-  /*
-   User wants to switch to another engine.
-   Only allow if there are no foreign key references to the table.
-  */
   if (is_child_or_parent_of_fk())
     DBUG_RETURN(0);
 
@@ -1692,8 +1793,8 @@ fk_split_name(char dst[], const char * src, bool index)
 
 struct Ndb_mem_root_guard {
   Ndb_mem_root_guard(MEM_ROOT *new_root) {
-    root_ptr= my_pthread_getspecific(MEM_ROOT**, THR_MALLOC);
-    DBUG_ASSERT(root_ptr != 0);
+    root_ptr= my_thread_get_THR_MALLOC();
+    assert(root_ptr != 0);
     old_root= *root_ptr;
     *root_ptr= new_root;
   }
@@ -1715,7 +1816,7 @@ ha_ndbcluster::get_fk_data(THD *thd, Ndb *ndb)
 
   free_root(mem_root, 0);
   m_fk_data= 0;
-  init_alloc_root(mem_root, fk_root_block_size, 0);
+  init_alloc_root(PSI_INSTRUMENT_ME, mem_root, fk_root_block_size, 0);
 
   NdbError err_OOM;
   err_OOM.code= 4000; // should we check OOM errors at all?
@@ -1786,7 +1887,7 @@ ha_ndbcluster::get_fk_data(THD *thd, Ndb *ndb)
       Ndb_table_guard child_tab(dict, child_name);
       if (child_tab.get_table() == 0)
       {
-        DBUG_ASSERT(false);
+        assert(false);
         ERR_RETURN(dict->getNdbError());
       }
 
@@ -1824,7 +1925,7 @@ ha_ndbcluster::get_fk_data(THD *thd, Ndb *ndb)
       Ndb_table_guard parent_tab(dict, parent_name);
       if (parent_tab.get_table() == 0)
       {
-        DBUG_ASSERT(false);
+        assert(false);
         ERR_RETURN(dict->getNdbError());
       }
 
@@ -1928,14 +2029,30 @@ ha_ndbcluster::get_fk_data(THD *thd, Ndb *ndb)
   }
 
   DBUG_PRINT("info", ("count FKs total %u child %u parent %u",
-                      data->list.elements, data->cnt_child, data->cnt_parent));
+                      data->list.elements, data->cnt_child,
+                      data->cnt_parent));
+
+  if (data->cnt_parent != 0 &&
+      global_system_variables.query_cache_type)
+  {
+    /*
+     Table has fk dependent child tables and query cache is enabled.
+     Build list of all dependent child tables that will
+     be updated during a cascading update/delete
+     */
+    res = data->build_cascading_dependents_list(ndb,
+                                                m_dbname, m_tabname,
+                                                &m_fk_mem_root);
+    if (res != 0)
+      DBUG_RETURN(res);
+  }
 
   m_fk_data= data;
   DBUG_RETURN(0);
 }
 
 void
-ha_ndbcluster::release_fk_data(THD *thd)
+ha_ndbcluster::release_fk_data()
 {
   DBUG_ENTER("ha_ndbcluster::release_fk_data");
 
@@ -1943,7 +2060,8 @@ ha_ndbcluster::release_fk_data(THD *thd)
   if (data != 0)
   {
     DBUG_PRINT("info", ("count FKs total %u child %u parent %u",
-                        data->list.elements, data->cnt_child, data->cnt_parent));
+                        data->list.elements, data->cnt_child,
+                        data->cnt_parent));
   }
 
   MEM_ROOT *mem_root= &m_fk_mem_root;
@@ -1964,7 +2082,7 @@ ha_ndbcluster::get_child_or_parent_fk_list(THD *thd,
   Ndb_fk_data *data= m_fk_data;
   if (data == 0)
   {
-    DBUG_ASSERT(false);
+    assert(false);
     DBUG_RETURN(0);
   }
 
@@ -2130,13 +2248,13 @@ ha_ndbcluster::get_foreign_key_create_info()
 
     fk_string.append(",");
     fk_string.append("\n ");
-    fk_string.append("CONSTRAINT `");
+    fk_string.append(" CONSTRAINT `");
     {
       char db_and_name[FN_LEN+1];
       const char * name = fk_split_name(db_and_name, fk.getName());
       fk_string.append(name);
     }
-    fk_string.append("` FOREIGN KEY(");
+    fk_string.append("` FOREIGN KEY (");
 
     {
       bool first= true;
@@ -2242,7 +2360,7 @@ errout:
 
     if (err.code != 0)
     {
-      push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
+      push_warning_printf(thd, Sql_condition::SL_WARNING,
                           ER_ILLEGAL_HA_CREATE_OPTION,
                           "Failed to retreive FK information: %d:%s",
                           err.code,
@@ -2275,8 +2393,8 @@ ha_ndbcluster::copy_fk_for_offline_alter(THD * thd, Ndb* ndb, NDBTAB* _dsttab)
   }
 
   Ndb_db_guard db_guard(ndb);
-  const char * src_db = thd->lex->select_lex.table_list.first->db;
-  const char * src_tab = thd->lex->select_lex.table_list.first->table_name;
+  const char * src_db = thd->lex->select_lex->table_list.first->db;
+  const char * src_tab = thd->lex->select_lex->table_list.first->table_name;
 
   if (src_db == 0 || src_tab == 0)
   {
@@ -2321,6 +2439,11 @@ ha_ndbcluster::copy_fk_for_offline_alter(THD * thd, Ndb* ndb, NDBTAB* _dsttab)
       bool found= false;
       for (unsigned i = 0; i < obj_list.count; i++)
       {
+        // Skip if the element is not a foreign key
+        if (obj_list.elements[i].type != NdbDictionary::Object::ForeignKey)
+          continue;
+
+        // Check if this is the fk being dropped
         char db_and_name[FN_LEN + 1];
         const char * name= fk_split_name(db_and_name,obj_list.elements[i].name);
         if (ndb_fk_casecmp(drop_item->name, name) != 0)
@@ -2329,9 +2452,16 @@ ha_ndbcluster::copy_fk_for_offline_alter(THD * thd, Ndb* ndb, NDBTAB* _dsttab)
         NdbDictionary::ForeignKey fk;
         if (dict->getForeignKey(fk, obj_list.elements[i].name) != 0)
         {
+          // should never happen
+          assert(false);
+          push_warning_printf(thd, Sql_condition::SL_WARNING,
+                              ER_CANT_DROP_FIELD_OR_KEY,
+                              "INTERNAL ERROR: Could not find foreign key '%s'",
+                              obj_list.elements[i].name);
           ERR_RETURN(dict->getNdbError());
         }
 
+        // The FK we are looking for is on src_tab.
         char child_db_and_name[FN_LEN + 1];
         const char* child_name = fk_split_name(child_db_and_name,
                                                fk.getChildTable());
@@ -2358,6 +2488,12 @@ ha_ndbcluster::copy_fk_for_offline_alter(THD * thd, Ndb* ndb, NDBTAB* _dsttab)
       NdbDictionary::ForeignKey fk;
       if (dict->getForeignKey(fk, obj_list.elements[i].name) != 0)
       {
+        // should never happen
+        assert(false);
+        push_warning_printf(thd, Sql_condition::SL_WARNING,
+                            ER_ALTER_INFO,
+                            "INTERNAL ERROR: Could not find foreign key '%s'",
+                            obj_list.elements[i].name);
         ERR_RETURN(dict->getNdbError());
       }
 
@@ -2466,7 +2602,24 @@ ha_ndbcluster::copy_fk_for_offline_alter(THD * thd, Ndb* ndb, NDBTAB* _dsttab)
         }
         else
         {
-          fk.setParent(* dsttab.get_table(), 0, cols);
+          /*
+            The parent column was previously the primary key.
+            Make sure it still is a primary key as implicit pks
+            might change during the alter. If not, get a better
+            matching index.
+           */
+          bool parent_primary = FALSE;
+          const NDBINDEX * idx = find_matching_index(dict,
+                                                     dsttab.get_table(),
+                                                     cols,
+                                                     parent_primary);
+          if (!parent_primary && idx == 0)
+          {
+            my_error(ER_FK_NO_INDEX_PARENT, MYF(0), fk.getName(),
+                     dsttab.get_table()->getName());
+            DBUG_RETURN(HA_ERR_CANNOT_ADD_FOREIGN);
+          }
+          fk.setParent(*dsttab.get_table(), idx, cols);
         }
 
 
@@ -2558,7 +2711,7 @@ ha_ndbcluster::drop_fk_for_online_alter(THD * thd, Ndb* ndb, NDBDICT * dict,
   Ndb_table_guard srctab(dict, tab->getName());
   if (srctab.get_table() == 0)
   {
-    DBUG_ASSERT(false); // Why ??
+    assert(false); // Why ??
     DBUG_RETURN(0);
   }
 
@@ -2607,6 +2760,15 @@ ha_ndbcluster::drop_fk_for_online_alter(THD * thd, Ndb* ndb, NDBDICT * dict,
         {
           ERR_RETURN(dict->getNdbError());
         }
+
+        /* Flush the parent table out if parent is different from child */
+        if(ndb_fk_casecmp(fk.getParentTable(), fk.getChildTable()) != 0)
+        {
+          char parent_db[FN_LEN + 1];
+          const char* parent_name = fk_split_name(parent_db,
+                                                  fk.getParentTable());
+          flush_parent_table_for_fk(thd, parent_db, parent_name);
+        }
         break;
       }
     }
@@ -2693,12 +2855,12 @@ ha_ndbcluster::recreate_fk_for_truncate(THD* thd, Ndb* ndb, const char* tab_name
   Ndb_table_guard child_tab(dict, tab_name);
   if (child_tab.get_table() == 0)
   {
-    push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
+    push_warning_printf(thd, Sql_condition::SL_WARNING,
                         ER_CANNOT_ADD_FOREIGN,
                         "INTERNAL ERROR: Could not find created child table '%s'",
                         tab_name);
     // Internal error, should be able to load the just created child table
-    DBUG_ASSERT(child_tab.get_table());
+    assert(child_tab.get_table());
     DBUG_RETURN(err_default);
   }
 
@@ -2718,11 +2880,11 @@ ha_ndbcluster::recreate_fk_for_truncate(THD* thd, Ndb* ndb, const char* tab_name
         const NDBCOL * ndbcol= tab->getColumn(fk->getChildColumnNo(i));
         if (ndbcol == 0)
         {
-          push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
+          push_warning_printf(thd, Sql_condition::SL_WARNING,
                               ER_CANNOT_ADD_FOREIGN,
                               "Child table %s has no column referred by the FK %s",
                               tab->getName(), fk->getName());
-          DBUG_ASSERT(ndbcol);
+          assert(ndbcol);
           DBUG_RETURN(err_default);
         }
         child_cols[pos++]= ndbcol;
@@ -2757,7 +2919,7 @@ ha_ndbcluster::recreate_fk_for_truncate(THD* thd, Ndb* ndb, const char* tab_name
       if (sscanf(fk->getName(), "%u/%u/%s",
                  &parent_id, &child_id, name) != 3)
       {
-        push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
+        push_warning_printf(thd, Sql_condition::SL_WARNING,
                             ER_CANNOT_ADD_FOREIGN,
                             "Skip, failed to parse name of fk: %s",
                             fk->getName());
@@ -2788,6 +2950,57 @@ ha_ndbcluster::recreate_fk_for_truncate(THD* thd, Ndb* ndb, const char* tab_name
     {
       ERR_RETURN(dict->getNdbError());
     }
+
+    /* Flush the parent table out if parent is different from child */
+    char parent_db[FN_LEN + 1];
+    const char* parent_name = fk_split_name(parent_db,
+                                            fk->getParentTable());
+    if(ndb_fk_casecmp(parent_name, tab_name) != 0 ||
+       ndb_fk_casecmp(parent_db, ndb->getDatabaseName()) != 0)
+    {
+      flush_parent_table_for_fk(thd, parent_db, parent_name);
+    }
   }
   DBUG_RETURN(0);
 }
+
+
+/**
+  @brief Append all the SHAREs of dependent children that would have been
+         updated in this transaction to the changed tables list
+
+  @param changed_tables        List to which the dependent children SHAREs
+                               has to be appended.
+  @param transaction_mem_root  Transaction's memroot
+  @return Void
+*/
+
+void
+ha_ndbcluster::append_dependents_to_changed_tables(
+    List<NDB_SHARE> &changed_tables,
+    MEM_ROOT* transaction_mem_root)
+{
+  DBUG_ENTER("ha_ndbcluster::append_dependents_to_changed_tables");
+
+  Ndb_fk_data *data= m_fk_data;
+
+  List<char> dependents_list = data->cascading_dependents;
+
+  const char* dependent_table;
+  List_iterator_fast<char> it(dependents_list);
+  while ((dependent_table= it++))
+  {
+    /* retrieve the NDB_SHARE using the key and add it to list */
+    char db_name[FN_REFLEN + 1];
+    const char* table_name = fk_split_name(db_name, dependent_table);
+    char key[FN_REFLEN + 1];
+    build_table_filename(key, sizeof(key) - 1, db_name,
+                         table_name, "", 0);
+    changed_tables.push_back(get_share(key, 0, TRUE),
+                             transaction_mem_root);
+  }
+
+  DBUG_PRINT("exit", ("Added %u dependents to list",
+                      dependents_list.elements));
+  DBUG_VOID_RETURN;
+}
