@@ -1,14 +1,21 @@
 /*
-   Copyright (c) 2000, 2010, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2000, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
@@ -20,12 +27,6 @@
   the NDB Cluster handler
 */
 
-#ifdef USE_PRAGMA_INTERFACE
-#pragma interface                       /* gcc class implementation */
-#endif
-
-#define round_up_byte(size) ((size + 7) >> 3) << 3
-
 typedef enum ndb_item_type {
   NDB_VALUE = 0,   // Qualified more with Item::Type
   NDB_FIELD = 1,   // Qualified from table definition
@@ -113,7 +114,7 @@ static const negated_function_mapping neg_map[]=
 class Ndb_item : public Sql_alloc
 {
 public:
-  Ndb_item(NDB_ITEM_TYPE item_type) : type(item_type) {};
+  Ndb_item(NDB_ITEM_TYPE item_type) : type(item_type) {}
   Ndb_item(NDB_ITEM_TYPE item_type, 
            NDB_ITEM_QUALIFICATION item_qualification,
            const Item *item_value)
@@ -138,7 +139,7 @@ public:
     case(NDB_END_COND):
       break;
     }
-  };
+  }
   Ndb_item(Field *field, int column_no) : type(NDB_FIELD)
   {
     NDB_ITEM_FIELD_VALUE *field_value= new NDB_ITEM_FIELD_VALUE();
@@ -146,20 +147,20 @@ public:
     field_value->field= field;
     field_value->column_no= column_no;
     value.field_value= field_value;
-  };
+  }
   Ndb_item(Item_func::Functype func_type, const Item *item_value) 
     : type(NDB_FUNCTION)
   {
     qualification.function_type= item_func_to_ndb_func(func_type);
     value.item= item_value;
     value.arg_count= ((Item_func *) item_value)->argument_count();
-  };
+  }
   Ndb_item(Item_func::Functype func_type, uint no_args) 
     : type(NDB_FUNCTION)
   {
     qualification.function_type= item_func_to_ndb_func(func_type);
     value.arg_count= no_args;
-  };
+  }
   ~Ndb_item()
   { 
     if (type == NDB_FIELD)
@@ -167,7 +168,7 @@ public:
         delete value.field_value;
         value.field_value= NULL;
       }
-  };
+  }
 
   uint32 pack_length() 
   { 
@@ -183,16 +184,16 @@ public:
     }
     
     return 0;
-  };
+  }
 
-  Field * get_field() { return value.field_value->field; };
+  Field * get_field() { return value.field_value->field; }
 
-  int get_field_no() { return value.field_value->column_no; };
+  int get_field_no() { return value.field_value->column_no; }
 
   int argument_count() 
   { 
     return value.arg_count;
-  };
+  }
 
   const char* get_val() 
   {  
@@ -203,34 +204,46 @@ public:
       break;
     case(NDB_FIELD):
       return (char*) value.field_value->field->ptr; 
+    case(NDB_FUNCTION):
+      if(qualification.value_type == Item::STRING_ITEM)
+        return value.item->str_value.ptr();
     default:
       break;
     }
     
     return NULL;
-  };
+  }
 
-  void save_in_field(Ndb_item *field_item)
+  const CHARSET_INFO *get_field_charset()
   {
+    Field *field= get_field();
+    if (field)
+      return field->charset();
+
+    return NULL;
+  }
+
+  const Item *get_item() const {
+    assert(this->type == NDB_VALUE);
+    return value.item;
+  }
+
+  uint32 save_in_field(Ndb_item *field_item)
+  {
+    uint32 length= 0;
     DBUG_ENTER("save_in_field");
     Field *field = field_item->value.field_value->field;
     const Item *item= value.item;
     if (item && field)
     {
-      DBUG_PRINT("info", ("item length %u, field length %u",
-                          item->max_length, field->field_length));
-      if (item->max_length > field->field_length)
-      {
-        DBUG_PRINT("info", ("Comparing field with longer value"));
-        DBUG_PRINT("info", ("Field can store %u", field->field_length));
-      }
+      length= item->max_length;
       my_bitmap_map *old_map=
         dbug_tmp_use_all_columns(field->table, field->table->write_set);
-      ((Item *)item)->save_in_field(field, FALSE);
+      ((Item *)item)->save_in_field(field, false);
       dbug_tmp_restore_column_map(field->table->write_set, old_map);
     }
-    DBUG_VOID_RETURN;
-  };
+    DBUG_RETURN(length);
+  }
 
   static NDB_FUNC_TYPE item_func_to_ndb_func(Item_func::Functype fun)
   {
@@ -251,14 +264,14 @@ public:
     case (Item_func::COND_OR_FUNC): { return NDB_COND_OR_FUNC; }
     default: { return NDB_UNSUPPORTED_FUNC; }
     }
-  };
+  }
 
   static NDB_FUNC_TYPE negate(NDB_FUNC_TYPE fun)
   {
     uint i= (uint) fun;
-    DBUG_ASSERT(fun == neg_map[i].pos_fun);
+    assert(fun == neg_map[i].pos_fun);
     return  neg_map[i].neg_fun;
-  };
+  }
 
   NDB_ITEM_TYPE type;
   NDB_ITEM_QUALIFICATION qualification;
@@ -273,7 +286,7 @@ public:
 class Ndb_cond : public Sql_alloc
 {
  public:
-  Ndb_cond() : ndb_item(NULL), next(NULL), prev(NULL) {};
+  Ndb_cond() : ndb_item(NULL), next(NULL), prev(NULL) {}
   ~Ndb_cond() 
   { 
     if (ndb_item) delete ndb_item; 
@@ -292,7 +305,7 @@ class Ndb_cond : public Sql_alloc
       delete tmp;
     }
     next= prev= NULL; 
-  };
+  }
   Ndb_item *ndb_item;
   Ndb_cond *next;
   Ndb_cond *prev;
@@ -308,14 +321,14 @@ class Ndb_cond : public Sql_alloc
 class Ndb_cond_stack : public Sql_alloc
 {
  public:
-  Ndb_cond_stack() : ndb_cond(NULL), next(NULL) {};
+  Ndb_cond_stack() : ndb_cond(NULL), next(NULL) {}
   ~Ndb_cond_stack() 
   { 
     if (ndb_cond) delete ndb_cond; 
     ndb_cond= NULL; 
     if (next) delete next;
     next= NULL; 
-  };
+  }
   Ndb_cond *ndb_cond;
   Ndb_cond_stack *next;
 };
@@ -332,22 +345,23 @@ class Ndb_cond_stack : public Sql_alloc
  */
 class Ndb_expect_stack : public Sql_alloc
 {
+  static const uint MAX_EXPECT_ITEMS = Item::VIEW_FIXER_ITEM + 1;
+  static const uint MAX_EXPECT_FIELD_TYPES = MYSQL_TYPE_GEOMETRY + 1;
+  static const uint MAX_EXPECT_FIELD_RESULTS = DECIMAL_RESULT + 1;
  public:
-Ndb_expect_stack(): collation(NULL), length(0), max_length(0), next(NULL) 
+  Ndb_expect_stack(): collation(NULL), length(0), max_length(0), next(NULL)
   {
-    // Allocate type checking bitmaps   
-    bitmap_init(&expect_mask,
-                0, round_up_byte(Item::MAX_NO_ITEMS), FALSE);
-    bitmap_init(&expect_field_type_mask,
-                0, round_up_byte(MAX_NO_FIELD_TYPES), FALSE);
-    bitmap_init(&expect_field_result_mask,
-                0, round_up_byte(MAX_NO_ITEM_RESULTS), FALSE);
-  };
+    // Allocate type checking bitmaps using fixed size buffers
+    // since max size is known at compile time
+    bitmap_init(&expect_mask, m_expect_buf,
+                MAX_EXPECT_ITEMS, FALSE);
+    bitmap_init(&expect_field_type_mask, m_expect_field_type_buf,
+                MAX_EXPECT_FIELD_TYPES, FALSE);
+    bitmap_init(&expect_field_result_mask, m_expect_field_result_buf,
+                MAX_EXPECT_FIELD_RESULTS, FALSE);
+  }
   ~Ndb_expect_stack()
   {
-    bitmap_free(&expect_mask);
-    bitmap_free(&expect_field_type_mask);
-    bitmap_free(&expect_field_result_mask);
     if (next)
       delete next;
     next= NULL;
@@ -384,6 +398,11 @@ Ndb_expect_stack(): collation(NULL), length(0), max_length(0), next(NULL)
   }
   bool expecting(Item::Type type)
   {
+    if (unlikely((uint)type > MAX_EXPECT_ITEMS))
+    {
+      // Unknown type, can't be expected
+      return false;
+    }
     return bitmap_is_set(&expect_mask, (uint) type);
   }
   void expect_nothing()
@@ -404,12 +423,21 @@ Ndb_expect_stack(): collation(NULL), length(0), max_length(0), next(NULL)
   {
     bitmap_set_bit(&expect_field_type_mask, (uint) type);
   }
+  void dont_expect_field_type(enum_field_types type)
+  {
+    bitmap_clear_bit(&expect_field_type_mask, (uint) type);
+  }
   void expect_all_field_types()
   {
     bitmap_set_all(&expect_field_type_mask);
   }
   bool expecting_field_type(enum_field_types type)
   {
+    if (unlikely((uint)type > MAX_EXPECT_FIELD_TYPES))
+    {
+      // Unknown type, can't be expected
+      return false;
+    }
     return bitmap_is_set(&expect_field_type_mask, (uint) type);
   }
   void expect_no_field_type()
@@ -432,6 +460,11 @@ Ndb_expect_stack(): collation(NULL), length(0), max_length(0), next(NULL)
   }
   bool expecting_field_result(Item_result result)
   {
+    if (unlikely((uint)result > MAX_EXPECT_FIELD_RESULTS))
+    {
+      // Unknown result, can't be expected
+      return false;
+    }
     return bitmap_is_set(&expect_field_result_mask,
                          (uint) result);
   }
@@ -448,11 +481,11 @@ Ndb_expect_stack(): collation(NULL), length(0), max_length(0), next(NULL)
     expect_no_field_result();
     expect_field_result(result);
   }
-  void expect_collation(CHARSET_INFO* col)
+  void expect_collation(const CHARSET_INFO* col)
   {
     collation= col;
   }
-  bool expecting_collation(CHARSET_INFO* col)
+  bool expecting_collation(const CHARSET_INFO* col)
   {
     bool matching= (!collation)
       ? true
@@ -483,10 +516,16 @@ Ndb_expect_stack(): collation(NULL), length(0), max_length(0), next(NULL)
   }
 
 private:
+  my_bitmap_map
+    m_expect_buf[bitmap_buffer_size(MAX_EXPECT_ITEMS)];
+  my_bitmap_map
+    m_expect_field_type_buf[bitmap_buffer_size(MAX_EXPECT_FIELD_TYPES)];
+  my_bitmap_map
+    m_expect_field_result_buf[bitmap_buffer_size(MAX_EXPECT_FIELD_RESULTS)];
   MY_BITMAP expect_mask;
   MY_BITMAP expect_field_type_mask;
   MY_BITMAP expect_field_result_mask;
-  CHARSET_INFO* collation;
+  const CHARSET_INFO* collation;
   Uint32 length;
   Uint32 max_length;
   Ndb_expect_stack* next;
@@ -496,7 +535,7 @@ class Ndb_rewrite_context : public Sql_alloc
 {
 public:
   Ndb_rewrite_context(Item_func *func) 
-    : func_item(func), left_hand_item(NULL), count(0) {};
+    : func_item(func), left_hand_item(NULL), count(0) {}
   ~Ndb_rewrite_context()
   {
     if (next) delete next;
@@ -559,6 +598,10 @@ class Ndb_cond_traverse_context : public Sql_alloc
   {
     expect_stack.expect_field_type(type);
   }
+  inline void dont_expect_field_type(enum_field_types type)
+  {
+    expect_stack.dont_expect_field_type(type);
+  }
   inline void expect_all_field_types()
   {
     expect_stack.expect_all_field_types();
@@ -600,11 +643,11 @@ class Ndb_cond_traverse_context : public Sql_alloc
   {
     expect_stack.expect_only_field_result(result);
   }
-  inline void expect_collation(CHARSET_INFO* col)
+  inline void expect_collation(const CHARSET_INFO* col)
   {
     expect_stack.expect_collation(col);
   }
-  inline bool expecting_collation(CHARSET_INFO* col)
+  inline bool expecting_collation(const CHARSET_INFO* col)
   {
     return expect_stack.expecting_collation(col);
   }
@@ -650,28 +693,28 @@ public:
   {}
   ~ha_ndbcluster_cond() 
   { if (m_cond_stack) delete m_cond_stack; }
-  const COND *cond_push(const COND *cond, 
+  const Item *cond_push(const Item *cond, 
                         TABLE *table, const NdbDictionary::Table *ndb_table);
   void cond_pop();
   void cond_clear();
   int generate_scan_filter(NdbInterpretedCode* code, 
-                           NdbScanOperation::ScanOptions* options);
-  int generate_scan_filter_from_cond(NdbScanFilter& filter);
+                           NdbScanOperation::ScanOptions* options) const;
+  int generate_scan_filter_from_cond(NdbScanFilter& filter) const;
   int generate_scan_filter_from_key(NdbInterpretedCode* code,
                                     NdbScanOperation::ScanOptions* options,
-                                    const KEY* key_info, 
+                                    const KEY* key_info,
                                     const key_range *start_key,
-                                    const key_range *end_key,
-                                    uchar *buf);
+                                    const key_range *end_key) const;
 private:
-  bool serialize_cond(const COND *cond, Ndb_cond_stack *ndb_cond,
-		      TABLE *table, const NdbDictionary::Table *ndb_table);
+  bool serialize_cond(const Item *cond, Ndb_cond_stack *ndb_cond,
+                      TABLE *table,
+                      const NdbDictionary::Table *ndb_table) const;
   int build_scan_filter_predicate(Ndb_cond* &cond, 
                                   NdbScanFilter* filter,
-                                  bool negated= false);
+                                  bool negated= false) const;
   int build_scan_filter_group(Ndb_cond* &cond, 
-                              NdbScanFilter* filter);
-  int build_scan_filter(Ndb_cond* &cond, NdbScanFilter* filter);
+                              NdbScanFilter* filter) const;
+  int build_scan_filter(Ndb_cond* &cond, NdbScanFilter* filter) const;
 
   Ndb_cond_stack *m_cond_stack;
 };
