@@ -1,14 +1,21 @@
 /*
-   Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
@@ -21,9 +28,12 @@
 #include "SignalData.hpp"
 #include <trigger_definitions.h>
 
+#define JAM_FILE_ID 27
+
+
 class LqhKeyReq {
   /**
-   * Reciver(s)
+   * Receiver(s)
    */
   friend class Dblqh;         // Reciver
 
@@ -71,9 +81,9 @@ private:
   union {
     /**
      * When sent from  TC -> LQH this variable contains scanInfo
-     * When send from LQH -> LQH this variable contains noFiredTriggers
+     * When send from LQH -> LQH this variable contains numFiredTriggers
      */
-    UintR noFiredTriggers;	// DATA 10 
+    UintR numFiredTriggers;	// DATA 10 
     Uint32 scanInfo;            // DATA 10
   };
 
@@ -134,6 +144,7 @@ private:
   static void setSeqNoReplica(UintR & requestInfo, UintR val);
   static void setLastReplicaNo(UintR & requestInfo, UintR val);
   static void setAIInLqhKeyReq(UintR & requestInfo, UintR val);
+  static void clearAIInLqhKeyReq(UintR & requestInfo);
   static void setKeyLen(UintR & requestInfo, UintR val);
   static void setSameClientAndTcFlag(UintR & requestInfo, UintR val);
   static void setReturnedReadLenAIFlag(UintR & requestInfo, UintR val);
@@ -169,10 +180,76 @@ private:
   static void setCorrFactorFlag(UintR & requestInfo, UintR val);
 
   /**
-   * Include corr factor
+   * Include deferred constraints
    */
   static UintR getDeferredConstraints(const UintR & requestInfo);
   static void setDeferredConstraints(UintR & requestInfo, UintR val);
+
+  /**
+   * Include disable foreign keys
+   */
+  static UintR getDisableFkConstraints(const UintR & requestInfo);
+  static void setDisableFkConstraints(UintR & requestInfo, UintR val);
+
+  /**
+   * Get mask of currently undefined bits
+   */
+  static UintR getLongClearBits(const UintR& requestInfo);
+  
+  /**
+   * Trigger flag ensuring that requests based on fully replicated triggers
+   * doesn't trigger a new trigger itself.
+   */
+  static UintR getNoTriggersFlag(const UintR & requestInfo);
+  static void setNoTriggersFlag(UintR & requestInfo, UintR val);
+
+  static UintR getUtilFlag (const UintR & requestInfo);
+  static void setUtilFlag(UintR & requestInfo, UintR val);
+
+  enum RequestInfo {
+    RI_KEYLEN_SHIFT      =  0, RI_KEYLEN_MASK      = 1023, /* legacy for short LQHKEYREQ */
+    RI_DISABLE_FK        =  0,
+    RI_NO_TRIGGERS       =  1,
+    RI_UTIL_SHIFT        =  2,
+
+    /* Currently unused */
+    RI_CLEAR_SHIFT3      =  3,
+    RI_CLEAR_SHIFT4      =  4,
+    RI_CLEAR_SHIFT5      =  5,
+    RI_CLEAR_SHIFT6      =  6,
+    RI_CLEAR_SHIFT7      =  7,
+    RI_CLEAR_SHIFT8      =  8,
+    RI_CLEAR_SHIFT9      =  9,
+
+    RI_LAST_REPL_SHIFT   = 10, RI_LAST_REPL_MASK   =    3,
+    RI_LOCK_TYPE_SHIFT   = 12, RI_LOCK_TYPE_MASK   =    7, /* legacy before ROWID_VERSION */
+    RI_GCI_SHIFT         = 12,
+    RI_NR_COPY_SHIFT     = 13,
+    RI_QUEUE_REDO_SHIFT  = 14,
+    RI_APPL_ADDR_SHIFT   = 15,
+    RI_DIRTY_SHIFT       = 16,
+    RI_INTERPRETED_SHIFT = 17,
+    RI_SIMPLE_SHIFT      = 18,
+    RI_OPERATION_SHIFT   = 19, RI_OPERATION_MASK   =    7,
+    RI_SEQ_REPLICA_SHIFT = 22, RI_SEQ_REPLICA_MASK =    3,
+    RI_AI_IN_THIS_SHIFT  = 24, RI_AI_IN_THIS_MASK  =    7, /* legacy for short LQHKEYREQ */
+    RI_CORR_FACTOR_VALUE = 24,
+    RI_NORMAL_DIRTY      = 25,
+    RI_DEFERRED_CONSTRAINTS = 26,
+    RI_SAME_CLIENT_SHIFT = 27,
+    RI_RETURN_AI_SHIFT   = 28,
+    RI_MARKER_SHIFT      = 29,
+    RI_NODISK_SHIFT      = 30,
+    RI_ROWID_SHIFT       = 31,
+  };
+
+  enum ScanInfo {
+   SI_ATTR_LEN_SHIFT    =  0, SI_ATTR_LEN_MASK  = 65535,
+   SI_STORED_PROC_SHIFT = 16,
+   SI_DISTR_KEY_SHIFT   = 17, SI_DISTR_KEY_MASK =   255,
+   SI_SCAN_TO_SHIFT     = 25,
+   SI_REORG_SHIFT       = 26, SI_REORG_MASK     =     3,
+  };
 };
 
 /**
@@ -203,6 +280,9 @@ private:
  * A = CorrFactor flag        - 1  Bit (24)
  * P = Do normal protocol even if dirty-read - 1 Bit (25)
  * D = Deferred constraints   - 1  Bit (26)
+ * F = Disable FK constraints - 1  Bit (0)
+ * T = no triggers            - 1  Bit (1)
+ * U = Operation came from UTIL - 1 Bit (2)
 
  * Short LQHKEYREQ :
  *             1111111111222222222233
@@ -213,37 +293,10 @@ private:
  * Long LQHKEYREQ :
  *             1111111111222222222233
  *   01234567890123456789012345678901
- *             llgnqpdisooorrAPDcumxz
+ *   FTU       llgnqpdisooorrAPDcumxz
  *
  */
 
-#define RI_KEYLEN_SHIFT      (0)
-#define RI_KEYLEN_MASK       (1023)
-#define RI_LAST_REPL_SHIFT   (10)
-#define RI_LAST_REPL_MASK    (3)
-#define RI_LOCK_TYPE_SHIFT   (12)
-#define RI_LOCK_TYPE_MASK    (7)
-#define RI_APPL_ADDR_SHIFT   (15)
-#define RI_DIRTY_SHIFT       (16)
-#define RI_INTERPRETED_SHIFT (17)
-#define RI_SIMPLE_SHIFT      (18)
-#define RI_OPERATION_SHIFT   (19)
-#define RI_OPERATION_MASK    (7)
-#define RI_SEQ_REPLICA_SHIFT (22)
-#define RI_SEQ_REPLICA_MASK  (3)
-#define RI_AI_IN_THIS_SHIFT  (24)
-#define RI_AI_IN_THIS_MASK   (7)
-#define RI_SAME_CLIENT_SHIFT (27)
-#define RI_RETURN_AI_SHIFT   (28)
-#define RI_MARKER_SHIFT      (29)
-#define RI_NODISK_SHIFT      (30)
-#define RI_ROWID_SHIFT       (31)
-#define RI_GCI_SHIFT         (12)
-#define RI_NR_COPY_SHIFT     (13)
-#define RI_QUEUE_REDO_SHIFT  (14)
-#define RI_CORR_FACTOR_VALUE (24)
-#define RI_NORMAL_DIRTY      (25)
-#define RI_DEFERRED_CONSTAINTS (26)
 
 /**
  * Scan Info
@@ -261,15 +314,6 @@ private:
  *                 pddddddddtmm       (Long LQHKEYREQ)
  */
 
-#define SI_ATTR_LEN_MASK     (65535)
-#define SI_ATTR_LEN_SHIFT    (0)
-#define SI_STORED_PROC_SHIFT (16)
-#define SI_DISTR_KEY_MASK    (255)
-#define SI_DISTR_KEY_SHIFT   (17)
-#define SI_SCAN_TO_SHIFT     (25)
-#define SI_REORG_SHIFT (26)
-#define SI_REORG_MASK  (3)
-
 inline 
 UintR
 LqhKeyReq::getAttrLen(const UintR & scanData)
@@ -512,6 +556,12 @@ LqhKeyReq::setAIInLqhKeyReq(UintR & requestInfo, UintR val){
   requestInfo |= (val << RI_AI_IN_THIS_SHIFT);
 }
 
+inline
+void
+LqhKeyReq::clearAIInLqhKeyReq(UintR & requestInfo){
+  requestInfo &= ~((Uint32)RI_AI_IN_THIS_MASK << RI_AI_IN_THIS_SHIFT);
+}
+
 inline
 void
 LqhKeyReq::setKeyLen(UintR & requestInfo, UintR val){
@@ -637,13 +687,68 @@ inline
 void
 LqhKeyReq::setDeferredConstraints(UintR & requestInfo, UintR val){
   ASSERT_BOOL(val, "LqhKeyReq::setDeferredConstraints");
-  requestInfo |= (val << RI_DEFERRED_CONSTAINTS);
+  requestInfo |= (val << RI_DEFERRED_CONSTRAINTS);
 }
 
 inline
 UintR
 LqhKeyReq::getDeferredConstraints(const UintR & requestInfo){
-  return (requestInfo >> RI_DEFERRED_CONSTAINTS) & 1;
+  return (requestInfo >> RI_DEFERRED_CONSTRAINTS) & 1;
+}
+
+inline
+void
+LqhKeyReq::setDisableFkConstraints(UintR & requestInfo, UintR val){
+  ASSERT_BOOL(val, "LqhKeyReq::setDisableFkConstraints");
+  requestInfo |= (val << RI_DISABLE_FK);
+}
+
+inline
+UintR
+LqhKeyReq::getDisableFkConstraints(const UintR & requestInfo){
+  return (requestInfo >> RI_DISABLE_FK) & 1;
+}
+
+inline
+UintR
+LqhKeyReq::getLongClearBits(const UintR& requestInfo)
+{
+  const Uint32 mask =
+    (1 << RI_CLEAR_SHIFT3) |
+    (1 << RI_CLEAR_SHIFT4) |
+    (1 << RI_CLEAR_SHIFT5) |
+    (1 << RI_CLEAR_SHIFT6) |
+    (1 << RI_CLEAR_SHIFT7) |
+    (1 << RI_CLEAR_SHIFT8) |
+    (1 << RI_CLEAR_SHIFT9);
+
+  return (requestInfo & mask);
+}
+
+inline
+void
+LqhKeyReq::setNoTriggersFlag(UintR & requestInfo, UintR val){
+  ASSERT_BOOL(val, "LqhKeyReq::setNoTriggersFlag");
+  requestInfo |= (val << RI_NO_TRIGGERS);
+}
+
+inline
+UintR
+LqhKeyReq::getNoTriggersFlag(const UintR & requestInfo){
+  return (requestInfo >> RI_NO_TRIGGERS) & 1;
+}
+
+inline
+void
+LqhKeyReq::setUtilFlag(UintR & requestInfo, UintR val){
+  ASSERT_BOOL(val, "LqhKeyReq::setUtilFlag");
+  requestInfo |= (val << RI_UTIL_SHIFT);
+}
+
+inline
+UintR
+LqhKeyReq::getUtilFlag(const UintR & requestInfo){
+  return (requestInfo >> RI_UTIL_SHIFT) & 1;
 }
 
 inline
@@ -710,7 +815,7 @@ private:
   };
   Uint32 transId1;
   Uint32 transId2;
-  Uint32 noFiredTriggers; // bit 31 defered trigger
+  Uint32 numFiredTriggers; // bit 31 defered trigger
 
   static Uint32 getFiredCount(Uint32 v) {
     return NoOfFiredTriggers::getFiredCount(v);
@@ -718,8 +823,14 @@ private:
   static Uint32 getDeferredUKBit(Uint32 v) {
     return NoOfFiredTriggers::getDeferredUKBit(v);
   }
-  static void setDeferredBit(Uint32 & v) {
-    NoOfFiredTriggers::setDeferredBit(v);
+  static void setDeferredUKBit(Uint32 & v) {
+    NoOfFiredTriggers::setDeferredUKBit(v);
+  }
+  static Uint32 getDeferredFKBit(Uint32 v) {
+    return NoOfFiredTriggers::getDeferredFKBit(v);
+  }
+  static void setDeferredFKBit(Uint32 & v) {
+    NoOfFiredTriggers::setDeferredFKBit(v);
   }
 };
 
@@ -756,4 +867,7 @@ private:
   Uint32 transId2;
 };
 
+
+#undef JAM_FILE_ID
+
 #endif
