@@ -1,17 +1,24 @@
-/* Copyright (c) 2008, 2014, Oracle and/or its affiliates. All rights reserved.
+/* Copyright (c) 2008, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA */
 
 #ifndef NDB_LOCAL_PROXY_HPP
 #define NDB_LOCAL_PROXY_HPP
@@ -30,6 +37,7 @@
 #include <signaldata/DropTrigImpl.hpp>
 #include <signaldata/DbinfoScan.hpp>
 #include <signaldata/Sync.hpp>
+#include <pgman.hpp>
 
 #define JAM_FILE_ID 438
 
@@ -493,9 +501,13 @@ protected:
   void sendDUMP_STATE_ORD(Signal*, Uint32 ssId, SectionHandle*);
 
   // GSN_NDB_TAMPER
-  struct Ss_NDB_TAMPER : SsParallel {
+  struct Ss_NDB_TAMPER : SsParallel
+  {
     Uint32 m_errorInsert;
-    Ss_NDB_TAMPER() {
+    Uint32 m_errorInsertExtra;
+    bool m_haveErrorInsertExtra;
+    Ss_NDB_TAMPER()
+    {
       m_sendREQ = &LocalProxy::sendNDB_TAMPER;
       m_sendCONF = 0;
     }
@@ -549,7 +561,7 @@ protected:
       m_sendREQ = &LocalProxy::sendDROP_TRIG_IMPL_REQ;
       m_sendCONF = &LocalProxy::sendDROP_TRIG_IMPL_CONF;
     }
-    enum { poolSize = 21 };
+    enum { poolSize = NDB_MAX_PROXY_DROP_TRIG_IMPL_REQ };
     static SsPool<Ss_DROP_TRIG_IMPL_REQ>& pool(LocalProxy* proxy) {
       return proxy->c_ss_DROP_TRIG_IMPL_REQ;
     }
