@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2003, 2013, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2.0,
@@ -33,6 +33,9 @@
 #include <Properties.hpp>
 #include <Configuration.hpp>
 
+#include <EventLogger.hpp>
+extern EventLogger * g_eventLogger;
+
 #define JAM_FILE_ID 472
 
 
@@ -52,6 +55,7 @@ Backup::Backup(Block_context& ctx, Uint32 instanceNumber) :
   addRecSignal(GSN_STTOR, &Backup::execSTTOR);
   addRecSignal(GSN_DUMP_STATE_ORD, &Backup::execDUMP_STATE_ORD);
   addRecSignal(GSN_READ_NODESCONF, &Backup::execREAD_NODESCONF);
+  addRecSignal(GSN_NODE_START_REP, &Backup::execNODE_START_REP, true);
   addRecSignal(GSN_NODE_FAILREP, &Backup::execNODE_FAILREP);
   addRecSignal(GSN_INCL_NODEREQ, &Backup::execINCL_NODEREQ);
   addRecSignal(GSN_CONTINUEB, &Backup::execCONTINUEB);
@@ -77,8 +81,6 @@ Backup::Backup(Block_context& ctx, Uint32 instanceNumber) :
   addRecSignal(GSN_DROP_TRIG_IMPL_CONF, &Backup::execDROP_TRIG_IMPL_CONF);
 
   addRecSignal(GSN_DIH_SCAN_TAB_CONF, &Backup::execDIH_SCAN_TAB_CONF);
-  addRecSignal(GSN_DIH_SCAN_GET_NODES_CONF,
-               &Backup::execDIH_SCAN_GET_NODES_CONF);
 
   addRecSignal(GSN_FSOPENREF, &Backup::execFSOPENREF, true);
   addRecSignal(GSN_FSOPENCONF, &Backup::execFSOPENCONF);
@@ -140,6 +142,9 @@ Backup::Backup(Block_context& ctx, Uint32 instanceNumber) :
   addRecSignal(GSN_END_LCPREQ, &Backup::execEND_LCPREQ);
 
   addRecSignal(GSN_DBINFO_SCANREQ, &Backup::execDBINFO_SCANREQ);
+
+  addRecSignal(GSN_CHECK_NODE_RESTARTCONF,
+               &Backup::execCHECK_NODE_RESTARTCONF);
 }
   
 Backup::~Backup()
@@ -158,25 +163,39 @@ Backup::execREAD_CONFIG_REQ(Signal* signal)
   Uint32 ref = req->senderRef;
   Uint32 senderData = req->senderData;
   ndbrequire(req->noOfParameters == 0);
+  jamEntry();
 
   const ndb_mgm_configuration_iterator * p = 
     m_ctx.m_config.getOwnConfigIterator();
   ndbrequire(p != 0);
 
-  c_defaults.m_disk_write_speed = 10 * (1024 * 1024);
-  c_defaults.m_disk_write_speed_sr = 100 * (1024 * 1024);
+  c_defaults.m_disk_write_speed_min = 10 * (1024 * 1024);
+  c_defaults.m_disk_write_speed_max = 20 * (1024 * 1024);
+  c_defaults.m_disk_write_speed_max_other_node_restart = 50 * (1024 * 1024);
+  c_defaults.m_disk_write_speed_max_own_restart = 100 * (1024 * 1024);
   c_defaults.m_disk_synch_size = 4 * (1024 * 1024);
   c_defaults.m_o_direct = true;
+  c_defaults.m_backup_disk_write_pct = 50;
 
   Uint32 noBackups = 0, noTables = 0, noAttribs = 0, noFrags = 0;
   ndbrequire(!ndb_mgm_get_int_parameter(p, CFG_DB_DISCLESS, 
 					&c_defaults.m_diskless));
   ndb_mgm_get_int_parameter(p, CFG_DB_O_DIRECT,
                             &c_defaults.m_o_direct);
-  ndb_mgm_get_int_parameter(p, CFG_DB_CHECKPOINT_SPEED_SR,
-			    &c_defaults.m_disk_write_speed_sr);
-  ndb_mgm_get_int_parameter(p, CFG_DB_CHECKPOINT_SPEED,
-			    &c_defaults.m_disk_write_speed);
+
+  ndb_mgm_get_int64_parameter(p, CFG_DB_MIN_DISK_WRITE_SPEED,
+			      &c_defaults.m_disk_write_speed_min);
+  ndb_mgm_get_int64_parameter(p, CFG_DB_MAX_DISK_WRITE_SPEED,
+			      &c_defaults.m_disk_write_speed_max);
+  ndb_mgm_get_int64_parameter(p,
+                CFG_DB_MAX_DISK_WRITE_SPEED_OTHER_NODE_RESTART,
+                &c_defaults.m_disk_write_speed_max_other_node_restart);
+  ndb_mgm_get_int64_parameter(p,
+                CFG_DB_MAX_DISK_WRITE_SPEED_OWN_RESTART,
+                &c_defaults.m_disk_write_speed_max_own_restart);
+  ndb_mgm_get_int_parameter(p, CFG_DB_BACKUP_DISK_WRITE_PCT,
+                            &c_defaults.m_backup_disk_write_pct);
+
   ndb_mgm_get_int_parameter(p, CFG_DB_DISK_SYNCH_SIZE,
 			    &c_defaults.m_disk_synch_size);
   ndb_mgm_get_int_parameter(p, CFG_DB_COMPRESSED_BACKUP,
@@ -184,16 +203,12 @@ Backup::execREAD_CONFIG_REQ(Signal* signal)
   ndb_mgm_get_int_parameter(p, CFG_DB_COMPRESSED_LCP,
 			    &c_defaults.m_compressed_lcp);
 
+  calculate_real_disk_write_speed_parameters();
+
+  jam();
   m_backup_report_frequency = 0;
   ndb_mgm_get_int_parameter(p, CFG_DB_BACKUP_REPORT_FREQUENCY, 
 			    &m_backup_report_frequency);
-  /*
-    We adjust the disk speed parameters from bytes per second to rather be
-    words per 100 milliseconds. We convert disk synch size from bytes per
-    second to words per second.
-  */
-  c_defaults.m_disk_write_speed /= (4 * 10);
-  c_defaults.m_disk_write_speed_sr /= (4 * 10);
 
   ndb_mgm_get_int_parameter(p, CFG_DB_PARALLEL_BACKUPS, &noBackups);
   //  ndbrequire(!ndb_mgm_get_int_parameter(p, CFG_DB_NO_TABLES, &noTables));
@@ -209,10 +224,28 @@ Backup::execREAD_CONFIG_REQ(Signal* signal)
   c_tablePool.setSize(noBackups * noTables + 1);
   c_triggerPool.setSize(noBackups * 3 * noTables);
   c_fragmentPool.setSize(noBackups * noFrags + 1);
-  
-  Uint32 szDataBuf = (2 * 1024 * 1024);
-  Uint32 szLogBuf = (2 * 1024 * 1024);
-  Uint32 szWrite = 32768, maxWriteSize = (256 * 1024);
+
+  c_tableMapSize = noBackups * noTables +1;
+  c_tableMap = (Uint32*)allocRecord("c_tableMap",
+                                    sizeof(Uint32),
+                                    c_tableMapSize);
+
+  for (Uint32 i = 0; i < c_tableMapSize; i++)
+  {
+    c_tableMap[i] = RNIL;
+  }
+
+  jam();
+
+  const Uint32 DEFAULT_WRITE_SIZE = (256 * 1024);
+  const Uint32 DEFAULT_MAX_WRITE_SIZE = (1024 * 1024);
+  const Uint32 DEFAULT_BUFFER_SIZE = (16 * 1024 * 1024);
+
+  Uint32 szDataBuf = DEFAULT_BUFFER_SIZE;
+  Uint32 szLogBuf = DEFAULT_BUFFER_SIZE;
+  Uint32 szWrite = DEFAULT_WRITE_SIZE;
+  Uint32 maxWriteSize = DEFAULT_MAX_WRITE_SIZE;
+
   ndb_mgm_get_int_parameter(p, CFG_DB_BACKUP_DATA_BUFFER_MEM, &szDataBuf);
   ndb_mgm_get_int_parameter(p, CFG_DB_BACKUP_LOG_BUFFER_MEM, &szLogBuf);
   ndb_mgm_get_int_parameter(p, CFG_DB_BACKUP_WRITE_SIZE, &szWrite);
@@ -234,8 +267,17 @@ Backup::execREAD_CONFIG_REQ(Signal* signal)
     maxWriteSize *= szWrite;
   }
 
+  /**
+   * Data buffer size must at least be big enough for a max-sized 
+   * scan batch.
+   */
+  ndbrequire(szDataBuf >= (BACKUP_MIN_BUFF_WORDS * 4));
+    
   /**
    * add min writesize to buffer size...and the alignment added here and there
+   * Need buffer size to be >= max-sized scan batch + min write size
+   * to avoid 'deadlock' where there's not enough buffered bytes to
+   * write, and too many bytes to fit another batch...
    */
   Uint32 extra = szWrite + 4 * (/* align * 512b */ 128);
 
@@ -248,25 +290,46 @@ Backup::execREAD_CONFIG_REQ(Signal* signal)
   c_defaults.m_maxWriteSize = maxWriteSize;
   c_defaults.m_lcp_buffer_size = szDataBuf;
 
+  /* We deprecate the use of BackupMemory, it serves no purpose at all. */
   Uint32 szMem = 0;
   ndb_mgm_get_int_parameter(p, CFG_DB_BACKUP_MEM, &szMem);
 
-  szMem += 3 * extra; // (data+log+lcp);
+  if (szMem != (32 * 1024 * 1024))
+  {
+    jam();
+    g_eventLogger->info("BackupMemory parameter setting ignored,"
+                        " BackupMemory deprecated");
+  }
+  szMem = szDataBuf + szLogBuf;
+
+  /**
+   * We allocate szDataBuf + szLogBuf pages for Backups and
+   * szDataBuf pages for LCPs.
+   */
   Uint32 noPages =
     (szMem + sizeof(Page32) - 1) / sizeof(Page32) +
     (c_defaults.m_lcp_buffer_size + sizeof(Page32) - 1) / sizeof(Page32);
 
   // We need to allocate an additional of 2 pages. 1 page because of a bug in
-  // ArrayPool and another one for DICTTAINFO.
+  // ArrayPool and another one for DICTTABINFO.
   c_pagePool.setSize(noPages + NO_OF_PAGES_META_FILE + 2, true); 
-  
+
+  jam();
+
   { // Init all tables
     SLList<Table> tables(c_tablePool);
     TablePtr ptr;
     while (tables.seizeFirst(ptr)){
       new (ptr.p) Table(c_fragmentPool);
+      ptr.p->backupPtrI = RNIL;
+      ptr.p->tableId = RNIL;
     }
-    while (tables.releaseFirst());
+    jam();
+    while (tables.releaseFirst())
+    {
+      ;
+    }
+    jam();
   }
 
   {
@@ -275,7 +338,12 @@ Backup::execREAD_CONFIG_REQ(Signal* signal)
     while (ops.seizeFirst(ptr)){
       new (ptr.p) BackupFile(* this, c_pagePool);
     }
-    while (ops.releaseFirst());
+    jam();
+    while (ops.releaseFirst())
+    {
+      ;
+    }
+    jam();
   }
   
   {
@@ -285,7 +353,12 @@ Backup::execREAD_CONFIG_REQ(Signal* signal)
       new (ptr.p) BackupRecord(* this, c_tablePool, 
 			       c_backupFilePool, c_triggerPool);
     }
-    while (recs.releaseFirst());
+    jam();
+    while (recs.releaseFirst())
+    {
+      ;
+    }
+    jam();
   }
 
   // Initialize BAT for interface to file system
@@ -307,3 +380,100 @@ Backup::execREAD_CONFIG_REQ(Signal* signal)
 	     ReadConfigConf::SignalLength, JBB);
 }
 
+/* Broken out in its own routine to enable setting via DUMP command. */
+void Backup::calculate_real_disk_write_speed_parameters(void)
+{
+  if (c_defaults.m_disk_write_speed_max < c_defaults.m_disk_write_speed_min)
+  {
+    /** 
+     * By setting max disk write speed equal or smaller than the minimum
+     * we will remove the adaptiveness of the LCP speed.
+     */
+    jam();
+    ndbout << "Setting MaxDiskWriteSpeed to MinDiskWriteSpeed since max < min"
+           << endl;
+    c_defaults.m_disk_write_speed_max = c_defaults.m_disk_write_speed_min;
+  }
+
+  if (c_defaults.m_disk_write_speed_max_other_node_restart <
+        c_defaults.m_disk_write_speed_max)
+  {
+    /** 
+     * By setting max disk write speed during restart equal or smaller than
+     * the maximum we will remove the extra adaptiveness of the LCP speed
+     * at other nodes restarts.
+     */
+    jam();
+    ndbout << "MaxDiskWriteSpeed larger than MaxDiskWriteSpeedOtherNodeRestart"
+           << " setting both to MaxDiskWriteSpeed" << endl;
+    c_defaults.m_disk_write_speed_max_other_node_restart =
+      c_defaults.m_disk_write_speed_max;
+  }
+
+  if (c_defaults.m_disk_write_speed_max_own_restart <
+        c_defaults.m_disk_write_speed_max_other_node_restart)
+  {
+    /** 
+     * By setting restart disk write speed during our restart equal or
+     * smaller than the maximum we will remove the extra adaptiveness of the
+     * LCP speed at other nodes restarts.
+     */
+    jam();
+    ndbout << "Setting MaxDiskWriteSpeedOwnRestart to "
+           << " MaxDiskWriteSpeedOtherNodeRestart since it was smaller"
+           << endl;
+    c_defaults.m_disk_write_speed_max_own_restart =
+      c_defaults.m_disk_write_speed_max_other_node_restart;
+  }
+
+  /*
+    We adjust the disk speed parameters from bytes per second to rather be
+    words per 100 milliseconds. We convert disk synch size from bytes per
+    second to words per second.
+  */
+  c_defaults.m_disk_write_speed_min /=
+    CURR_DISK_SPEED_CONVERSION_FACTOR_TO_SECONDS;
+  c_defaults.m_disk_write_speed_max /=
+    CURR_DISK_SPEED_CONVERSION_FACTOR_TO_SECONDS;
+  c_defaults.m_disk_write_speed_max_other_node_restart /=
+    CURR_DISK_SPEED_CONVERSION_FACTOR_TO_SECONDS;
+  c_defaults.m_disk_write_speed_max_own_restart /=
+    CURR_DISK_SPEED_CONVERSION_FACTOR_TO_SECONDS;
+
+  Uint32 num_ldm_threads = globalData.ndbMtLqhThreads;
+  if (num_ldm_threads == 0)
+  {
+    /* We are running with ndbd binary */
+    jam();
+    num_ldm_threads = 1;
+  }
+  c_defaults.m_disk_write_speed_min /= num_ldm_threads;
+  c_defaults.m_disk_write_speed_max /= num_ldm_threads;
+  c_defaults.m_disk_write_speed_max_other_node_restart /= num_ldm_threads;
+  c_defaults.m_disk_write_speed_max_own_restart /= num_ldm_threads;
+}
+
+void Backup::restore_disk_write_speed_numbers(void)
+{
+  c_defaults.m_disk_write_speed_min *=
+    CURR_DISK_SPEED_CONVERSION_FACTOR_TO_SECONDS;
+  c_defaults.m_disk_write_speed_max *=
+    CURR_DISK_SPEED_CONVERSION_FACTOR_TO_SECONDS;
+  c_defaults.m_disk_write_speed_max_other_node_restart *=
+    CURR_DISK_SPEED_CONVERSION_FACTOR_TO_SECONDS;
+  c_defaults.m_disk_write_speed_max_own_restart *=
+    CURR_DISK_SPEED_CONVERSION_FACTOR_TO_SECONDS;
+
+  Uint32 num_ldm_threads = globalData.ndbMtLqhThreads;
+  if (num_ldm_threads == 0)
+  {
+    /* We are running with ndbd binary */
+    jam();
+    num_ldm_threads = 1;
+  }
+
+  c_defaults.m_disk_write_speed_min *= num_ldm_threads;
+  c_defaults.m_disk_write_speed_max *= num_ldm_threads;
+  c_defaults.m_disk_write_speed_max_other_node_restart *= num_ldm_threads;
+  c_defaults.m_disk_write_speed_max_own_restart *= num_ldm_threads;
+}
