@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2005, 2015, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2005, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2.0,
@@ -38,6 +38,16 @@
 
 #define JAM_FILE_ID 462
 
+/**
+ * ndbd has only one pgman sharing the single disk page buffer between
+ * the locked extent pages and the dynamic caching of data pages from
+ * data files.  Limit the extent page usage to 25% of the buffer. One
+ * extent page can describe 1022 extents.  25% of the default
+ * DiskPageBufferMemory of 64 MB (512 pages) will be sufficient to
+ * open a single data file of max recommended size 32GB, needing 33
+ * extent pages.
+ */
+#define NDBD_EXTENT_PAGE_PERCENT 25
 
 /*
  * PGMAN
@@ -249,9 +259,12 @@ class Pgman : public SimulatedBlock
 public:
   Pgman(Block_context& ctx, Uint32 instanceNumber = 0);
   virtual ~Pgman();
+  /* Special function to indicate the block is the extra PGMAN worker */
+  void init_extra_pgman();
   BLOCK_DEFINES(Pgman);
 
 private:
+  friend class Tsman;
   friend class Page_cache_client;
   friend class PgmanProxy;
 
@@ -379,6 +392,24 @@ private:
   typedef DLCFifoList<Page_entry, Page_entry_queue_ptr> Page_queue;
   typedef DLCFifoList<Page_entry, Page_entry_sublist_ptr> Page_sublist;
 
+  /**
+   * In ndbmtd, there is an extra pgman instance not associated with
+   * an LDM, which is mostly used for accessing extent pages
+   * containing metadata about extents.
+   */
+  bool m_extra_pgman;
+  /**
+   * extra_pgman reserves 'm_extra_pgman_reserve' slots of the disk
+   * page buffer memory for undo log execution during restart. These
+   * will be used to read in data pages in order to find the tableid
+   * and fragmentid of an undo log record (in order to decide the ldm
+   * that will undo this log record), if they cannot be retrieved from
+   * tsman extent info.  These pages can be evicted after retrieving
+   * table/fragids, and not given to the ldm for undo. The ldm will
+   * read the page itself into its share of the disk page buffer memory.
+   */
+  Uint32 m_extra_pgman_reserve_pages;
+
   class Dbtup *c_tup;
   class Lgman *c_lgman;
   class Tsman *c_tsman;
@@ -446,7 +477,8 @@ private:
       m_page_requests_direct_return(0),
       m_page_requests_wait_q(0),
       m_page_requests_wait_io(0),
-      m_entries_high(0)
+      m_entries_high(0),
+      m_num_locked_pages(0)
     {}
     Uint32 m_num_pages;         // current number of cache pages
     Uint32 m_num_hot_pages;
@@ -461,6 +493,7 @@ private:
     Uint64 m_page_requests_wait_q;
     Uint64 m_page_requests_wait_io;
     Uint32 m_entries_high;
+    Uint32 m_num_locked_pages;
   } m_stats;
 
   enum CallbackIndex {
@@ -548,6 +581,7 @@ private:
   void map_file_no(Uint32 file_no, Uint32 fd);
   void free_data_file(Uint32 file_no, Uint32 fd = RNIL);
   int drop_page(Ptr<Page_entry>);
+  bool extent_pages_available(Uint32 pages_needed);
   
 #ifdef VM_TRACE
   bool debugFlag;        // not yet in use in 7.0
@@ -621,7 +655,13 @@ public:
    *         >0 is ok
    */
   int drop_page(Local_key, Uint32 page_id);
-  
+
+  /**
+   * Check whether there are 'pages_needed' pages available
+   * to be used as extent pages by the extra_pgman.
+   */
+  bool extent_pages_available(Uint32 pages_needed);
+
   /**
    * Create file record
    */
