@@ -1,14 +1,21 @@
 /*
-   Copyright (c) 2003, 2014, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
@@ -24,6 +31,7 @@
 #include <SimulatedBlock.hpp>
 #include <NodeBitmask.hpp>
 #include <SignalCounter.hpp>
+#include <ProcessInfo.hpp>
 
 #include <signaldata/EventReport.hpp>
 #include <signaldata/ArbitSignalData.hpp>
@@ -371,6 +379,9 @@ private:
   // Ndbinfo signal
   void execDBINFO_SCANREQ(Signal *signal);
 
+  // ProcessInfo Report signal
+  void execPROCESSINFO_REP(Signal *signal);
+
   // NDBCNTR informing us our node is fully started
   void execNODE_STARTED_REP(Signal *signal);
 
@@ -418,7 +429,7 @@ private:
                      Uint16 sourceNode);
   void sendCommitFailReq(Signal* signal);
   void presToConfLab(Signal* signal);
-  void sendSttorryLab(Signal* signal);
+  void sendSttorryLab(Signal* signal, bool first_phase);
   void sttor020Lab(Signal* signal);
   void closeComConfLab(Signal* signal);
   void apiRegReqLab(Signal* signal);
@@ -482,6 +493,7 @@ private:
   void joinedCluster(Signal* signal, NodeRecPtr nodePtr);
   void sendCmRegReq(Signal * signal, Uint32 nodeId);
   void sendCmNodeInfoReq(Signal* signal, Uint32 nodeId, const NodeRec * self);
+  ProcessInfo * getProcessInfo(Uint32 nodeId);
 
 private:
   void sendPrepFailReqRef(Signal* signal, 
@@ -542,6 +554,9 @@ private:
   Timer hb_send_timer;
   Timer hb_api_timer;
 
+  Int16 processInfoNodeIndex[MAX_NODES];
+  ProcessInfo * receivedProcessInfo;
+  Uint16 max_api_node_id;
 
   NdbNodeBitmask cfailedNodes;
   NdbNodeBitmask cprepFailedNodes;
