@@ -1,5 +1,5 @@
 /*
-   Copyright (c) 2003, 2014, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License, version 2.0,
@@ -31,7 +31,6 @@
 #include <pc.hpp>
 #include "Pool.hpp"
 #include <ErrorReporter.hpp>
-#include <NdbMem.h>
 #include <Bitmask.hpp>
 #include <mgmapi.h>
 
@@ -164,13 +163,19 @@ public:
 
   /**
    * Return an object to pool
+   * release releases the object and places it first in free list
+   * releaseLast releases the object and places it last in free list
    */
   void release(Uint32 i);
+  void releaseLast(Uint32 i);
 
   /**
    * Return an object to pool
+   * release releases the object and places it first in free list
+   * releaseLast releases the object and places it last in free list
    */
   void release(Ptr<T> &);
+  void releaseLast(Ptr<T> &);
 
 #ifdef ARRAY_GUARD
   /**
@@ -194,7 +199,8 @@ public:
    */
   struct Cache
   {
-    Cache(Uint32 a0 = 512, Uint32 a1 = 256) { m_first_free = RNIL; m_free_cnt = 0; m_alloc_cnt = a0; m_max_free_cnt = a1; }
+    Cache(Uint32 a0 = 512, Uint32 a1 = 256)
+    { m_first_free = RNIL; m_free_cnt = 0; m_alloc_cnt = a0; m_max_free_cnt = a1; }
     void init_cache(Uint32 a0, Uint32 a1)
     {
       m_alloc_cnt = a0;
@@ -292,6 +298,7 @@ protected:
    */
   char protect_read_var[NDB_CL_PADSZ(sizeof(Uint32) + sizeof(void*))];
   Uint32 firstFree;
+  Uint32 lastFree;
   Uint32 noOfFree;
   Uint32 noOfFreeMin;
 #ifdef ARRAY_GUARD
@@ -310,6 +317,7 @@ ArrayPool<T>::ArrayPool(CallBack* seizeErrorHandler):
   seizeErrHand(seizeErrorHandler)
 {
   firstFree = RNIL;
+  lastFree = RNIL;
   size = 0;
   noOfFree = 0;
   noOfFreeMin = 0;
@@ -346,18 +354,23 @@ inline
 bool
 ArrayPool<T>::setSize(Uint32 noOfElements, 
 		      bool align, bool exit_on_error, bool guard, Uint32 paramId){
-  if(size == 0){
+  if(size == 0)
+  {
     if(noOfElements == 0)
+    {
       return true;
-  Uint64 real_size = (Uint64)noOfElements * sizeof(T);
-  size_t req_size = (size_t)real_size;
-  Uint64 real_size_align = real_size + sizeof(T);
-  size_t req_size_align = (size_t)real_size_align;
+    }
+    Uint64 real_size = (Uint64)noOfElements * sizeof(T);
+    size_t req_size = (size_t)real_size;
+    Uint64 real_size_align = real_size + sizeof(T);
+    size_t req_size_align = (size_t)real_size_align;
 
     if(align)
     {
       if((Uint64)req_size_align == real_size_align && req_size_align > 0)
-        alloc_ptr = ndbd_malloc(req_size_align);  
+      {
+        alloc_ptr = ndbd_malloc(req_size_align);
+      }
       UintPtr p = (UintPtr)alloc_ptr;
       UintPtr mod = p % sizeof(T);
       if (mod)
@@ -367,7 +380,9 @@ ArrayPool<T>::setSize(Uint32 noOfElements,
       theArray = (T *)p;
     }
     else if((Uint64)req_size == real_size && req_size > 0)
+    {
       theArray = (T *)(alloc_ptr = ndbd_malloc(req_size));
+    }
 
     if(theArray == 0)
     {
@@ -377,7 +392,9 @@ ArrayPool<T>::setSize(Uint32 noOfElements,
       if (!exit_on_error)
 	return false;
 
-      if(0 != paramId && 0 == ndb_mgm_get_db_parameter_info(paramId, &param_info, &tsize)) {
+      if(0 != paramId &&
+         0 == ndb_mgm_get_db_parameter_info(paramId, &param_info, &tsize))
+      {
         BaseString::snprintf(errmsg, sizeof(errmsg), 
                 "Malloc memory for %s failed", param_info.m_name);
       }
@@ -394,12 +411,14 @@ ArrayPool<T>::setSize(Uint32 noOfElements,
      * Set next pointers
      */
     T * t = &theArray[0];
-    for(Uint32 i = 0; i<size; i++){
+    for(Uint32 i = 0; i<size; i++)
+    {
       t->nextPool = (i + 1);
       t++;
     }
     theArray[size-1].nextPool = RNIL;
     firstFree = 0;
+    lastFree = size - 1;
 
 #ifdef ARRAY_GUARD
     if (guard)
@@ -413,7 +432,9 @@ ArrayPool<T>::setSize(Uint32 noOfElements,
     return true;
   }
   if (!exit_on_error)
+  {
     return false;
+  }
 
   ErrorReporter::handleAssert("ArrayPool<T>::setSize called twice", __FILE__, __LINE__);
   return false; // not reached
@@ -782,6 +803,7 @@ ArrayPool<T>::getPtrIgnoreAlloc(Ptr<T> & ptr){
  *
  * Return i
  */
+
 template <class T>
 inline
 bool
@@ -796,6 +818,11 @@ ArrayPool<T>::seize(Ptr<T> & ptr){
     
     ptr.i = ff;
     ptr.p = &theArray[ff];
+    if (firstFree == RNIL)
+    {
+      assert(lastFree == ff);
+      lastFree = RNIL;
+    }
 #ifdef ARRAY_GUARD
     if (theAllocatedBitmask)
     {
@@ -844,6 +871,11 @@ ArrayPool<T>::seizeId(Ptr<T> & ptr, Uint32 i){
       firstFree = theArray[ff].nextPool;
     else
       theArray[prev].nextPool = theArray[ff].nextPool;
+    if (ff == lastFree)
+    {
+      assert(firstFree == RNIL);
+      lastFree = prev;
+    }
     
     ptr.i = ff;
     ptr.p = &theArray[ff];
@@ -1040,6 +1072,11 @@ ArrayPool<T>::release(Uint32 _i){
     Uint32 ff = firstFree;
     theArray[i].nextPool = ff;
     firstFree = i;
+    if (ff == RNIL)
+    {
+      assert(lastFree == RNIL);
+      lastFree = i;
+    }
 
 #ifdef ARRAY_GUARD
     if (theAllocatedBitmask)
@@ -1061,6 +1098,56 @@ ArrayPool<T>::release(Uint32 _i){
   ErrorReporter::handleAssert("ArrayPool<T>::release", __FILE__, __LINE__);
 }
 
+template <class T>
+inline
+void
+ArrayPool<T>::releaseLast(Uint32 _i){
+
+#ifdef ARRAY_GUARD
+  assert(chunk == false);
+#endif
+
+  const Uint32 i = _i;
+  if(likely(i < size))
+  {
+    Uint32 lf = lastFree;
+    lastFree = i;
+    theArray[i].nextPool = RNIL;
+    if (lf == RNIL)
+    {
+      assert(firstFree == RNIL);
+      firstFree = i;
+    }
+    else if (lf < size)
+    {
+      theArray[lf].nextPool = i;
+    }
+    else
+    {
+      goto error;
+    }
+
+#ifdef ARRAY_GUARD
+    if (theAllocatedBitmask)
+    {
+      if(BitmaskImpl::get(bitmaskSz, theAllocatedBitmask, i)){
+	BitmaskImpl::clear(bitmaskSz, theAllocatedBitmask, i);
+	noOfFree++;
+	return;
+      }
+      /**
+       * Relesing a already released element
+       */
+      ErrorReporter::handleAssert("ArrayPool<T>::releaseLast", __FILE__, __LINE__);
+    }
+#endif
+    noOfFree++;
+    return;
+  }
+error:
+  ErrorReporter::handleAssert("ArrayPool<T>::release", __FILE__, __LINE__);
+}
+
 /**
  * Return an object to pool
  */
@@ -1078,6 +1165,11 @@ ArrayPool<T>::release(Ptr<T> & ptr){
     Uint32 ff = firstFree;
     theArray[i].nextPool = ff;
     firstFree = i;
+    if (ff == RNIL)
+    {
+      assert(lastFree == RNIL);
+      lastFree = i;
+    }
 
 #ifdef ARRAY_GUARD
     if (theAllocatedBitmask)
@@ -1100,6 +1192,58 @@ ArrayPool<T>::release(Ptr<T> & ptr){
   ErrorReporter::handleAssert("ArrayPool<T>::release", __FILE__, __LINE__);
 }
 
+template <class T>
+inline
+void
+ArrayPool<T>::releaseLast(Ptr<T> & ptr)
+{
+
+#ifdef ARRAY_GUARD
+  assert(chunk == false);
+#endif
+
+  Uint32 i = ptr.i;
+  Uint32 lf = lastFree;
+  if(likely(i < size))
+  {
+    lastFree = i;
+    theArray[i].nextPool = RNIL;
+    if (lf < size)
+    {
+      theArray[lf].nextPool = i;
+    }
+    else if (lf == RNIL)
+    {
+      assert(firstFree == RNIL);
+      firstFree = i;
+    }
+    else
+    {
+      goto error;
+    }
+
+#ifdef ARRAY_GUARD
+    if (theAllocatedBitmask)
+    {
+      if(BitmaskImpl::get(bitmaskSz, theAllocatedBitmask, i)){
+	BitmaskImpl::clear(bitmaskSz, theAllocatedBitmask, i);
+	//assert(noOfFree() == noOfFree2());
+	noOfFree++;
+	return;
+      }
+      /**
+       * Relesing a already released element
+       */
+      ErrorReporter::handleAssert("ArrayPool<T>::releaseLast", __FILE__, __LINE__);
+    }
+#endif
+    noOfFree++;
+    return;
+  }
+error:
+  ErrorReporter::handleAssert("ArrayPool<T>::release", __FILE__, __LINE__);
+}
+
 #if 0
 #define DUMP(a,b) do { printf("%s c.m_first_free: %u c.m_free_cnt: %u %s", a, c.m_first_free, c.m_free_cnt, b); fflush(stdout); } while(0)
 #else
@@ -1554,7 +1698,7 @@ void
 SafeArrayPool<T>::release(Uint32 i)
 {
   int ret = lock();
-  assert(ret == 0);
+  require(ret == 0);
   ArrayPool<T>::release(i);
   unlock();
 }
@@ -1565,7 +1709,7 @@ void
 SafeArrayPool<T>::release(Ptr<T>& ptr)
 {
   int ret = lock();
-  assert(ret == 0);
+  require(ret == 0);
   ArrayPool<T>::release(ptr);
   unlock();
 }
