@@ -1,14 +1,21 @@
 /*
-   Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
@@ -26,6 +33,7 @@
 #include <m_string.h>
 #include <util/SparseBitmask.hpp>
 #include "../common/util/parse_mask.hpp"
+#include <stdlib.h>
 
 extern EventLogger *g_eventLogger;
 
@@ -299,6 +307,10 @@ InitConfigFileParser::storeNameValuePair(Context& ctx,
       ctx.reportWarning("[%s] %s is deprecated", ctx.fname, fname);
     }
   }
+  if( status == ConfigInfo::CI_INTERNAL) {
+    ctx.reportError("[%s] Parameter %s not configurable by user", ctx.fname, fname);
+    return false;
+  }
 
   const ConfigInfo::Type type = m_info->getType(ctx.m_currentInfo, fname);
   switch(type){
@@ -372,6 +384,9 @@ InitConfigFileParser::storeNameValuePair(Context& ctx,
       case -2:
         desc.assfmt("Too large id used in bitmask, max is %llu", max);
         break;
+      case -3:
+        desc.assfmt("Empty bitmask not allowed");
+        break;
       default:
         break;
       }
@@ -411,42 +426,78 @@ bool InitConfigFileParser::isEmptyLine(const char* line) const {
 //  Convert String to Int
 //****************************************************************************
 bool InitConfigFileParser::convertStringToUint64(const char* s, 
-						 Uint64& val,
-						 Uint32 log10base) {
+                                                 Uint64& val)
+{
   if (s == NULL)
     return false;
-  if (strlen(s) == 0) 
-    return false;
 
-  errno = 0;
+  while (*s != '\0' && isspace(*s)) s++;
+  const bool negative = (*s == '-');
+
+  union {
+    signed long long vs;
+    unsigned long long vu;
+  };
   char* p;
-  Int64 v = strtoll(s, &p, log10base);
-  if (errno != 0)
-    return false;
-  
-  long mul = 0;
-  if (p != &s[strlen(s)]){
-    char * tmp = strdup(p);
-    trim(tmp);
-    switch(tmp[0]){
-    case 'k':
-    case 'K':
-      mul = 10;
-      break;
-    case 'M':
-      mul = 20;
-      break;
-    case 'G':
-      mul = 30;
-      break;
-    default:
-      free(tmp);
+  const int log10base = 0;
+  errno = 0;
+  if (negative)
+  {
+    vs = strtoll(s, &p, log10base);
+    if ((vs == LLONG_MIN || vs == LLONG_MAX) && errno == ERANGE)
+    {
       return false;
     }
-    free(tmp);
   }
-  
-  val = (v << mul);
+  else
+  {
+    vu = strtoull(s, &p, log10base);
+    if (vu == ULLONG_MAX && errno == ERANGE)
+    {
+      return false;
+    }
+  }
+  if (p == s)
+    return false;
+
+  int mul = 0;
+
+  switch(*p)
+  {
+  case '\0':
+    break;
+  case 'k':
+  case 'K':
+    mul = 10;
+    p++;
+    break;
+  case 'M':
+    mul = 20;
+    p++;
+    break;
+  case 'G':
+    mul = 30;
+    p++;
+    break;
+  default:
+    return false;
+  }
+  if (*p != '\0')
+    return false;
+  if (negative)
+  {
+    Int64 v = (vs << mul);
+    if ((v >> mul) != vs)
+      return false;
+    val = v;
+  }
+  else
+  {
+    Uint64 v = (vu << mul);
+    if ((v >> mul) != vu)
+      return false;
+    val = v;
+  }
   return true;
 }
 
@@ -548,10 +599,12 @@ InitConfigFileParser::parseDefaultSectionHeader(const char* line) const {
   int no = sscanf(line, "[%120[A-Z_a-z] %120[A-Z_a-z]]", token1, token2);
 
   // Not correct no of tokens 
-  if (no != 2) return NULL;
+  if (no != 2)
+    return NULL;
 
   // Not correct keyword at end
-  if (!strcasecmp(token2, "DEFAULT") == 0) return NULL;
+  if (native_strcasecmp(token2, "DEFAULT") != 0)
+    return NULL;
 
   const char *token1_alias= m_info->getAlias(token1);
   if (token1_alias == 0)
@@ -645,9 +698,7 @@ InitConfigFileParser::Context::reportWarning(const char * fmt, ...){
 
 #include <my_sys.h>
 #include <my_getopt.h>
-#ifdef HAVE_MY_DEFAULT_H
 #include <my_default.h>
-#endif
 
 static int order = 1;
 static 
@@ -740,10 +791,7 @@ load_defaults(Vector<struct my_option>& options, const char* groups[])
   BaseString group_suffix;
 
   const char *save_file = my_defaults_file;
-#if MYSQL_VERSION_ID >= 50508
-  const
-#endif
-  char *save_extra_file = my_defaults_extra_file;
+  const char *save_extra_file = my_defaults_extra_file;
   const char *save_group_suffix = my_defaults_group_suffix;
 
   if (my_defaults_file)
@@ -798,7 +846,7 @@ InitConfigFileParser::load_mycnf_groups(Vector<struct my_option> & options,
   }
 
   struct my_option end;
-  bzero(&end, sizeof(end));
+  memset(&end, 0, sizeof(end));
   copy.push_back(end);
 
   if (load_defaults(copy, groups))
@@ -816,7 +864,7 @@ InitConfigFileParser::parse_mycnf()
   {
     {
       struct my_option opt;
-      bzero(&opt, sizeof(opt));
+      memset(&opt, 0, sizeof(opt));
       const ConfigInfo::ParamInfo& param = ConfigInfo::m_ParamInfo[i];
       switch(param._type){
       case ConfigInfo::CI_BOOL:
@@ -857,7 +905,7 @@ InitConfigFileParser::parse_mycnf()
   Uint32 idx = options.size();
   {
     struct my_option opt;
-    bzero(&opt, sizeof(opt));
+    memset(&opt, 0, sizeof(opt));
     opt.name = "ndbd";
     opt.id = 256;
     opt.value = (uchar**)malloc(sizeof(char*));
@@ -886,7 +934,7 @@ InitConfigFileParser::parse_mycnf()
     opt.arg_type = REQUIRED_ARG;
     options.push_back(opt);
 
-    bzero(&opt, sizeof(opt));
+    memset(&opt, 0, sizeof(opt));
     options.push_back(opt);
 
     ndbd = &options[idx];
@@ -1026,3 +1074,58 @@ template class Vector<struct my_option>;
 /*
   See include/my_getopt.h for the declaration of struct my_option
 */
+
+
+#ifdef TEST_INITCONFIGFILEPARSER
+
+#include "../../../../unittest/mytap/tap.h"
+
+int main()
+{
+  plan(29);
+
+  // Check the string to int parsing function convertStringToUint64()
+  Uint64 value;
+  ok1(InitConfigFileParser::convertStringToUint64("37", value));
+  ok1(value == 37);
+  ok1(InitConfigFileParser::convertStringToUint64("0", value));
+  ok1(value == 0);
+  ok1(InitConfigFileParser::convertStringToUint64("-0", value)); // ??
+  ok1(static_cast<Int64>(value) == -0);
+  ok1(InitConfigFileParser::convertStringToUint64("-1", value));
+  ok1(static_cast<Int64>(value) == -1);
+  ok1(InitConfigFileParser::convertStringToUint64("-37", value));
+  ok1(static_cast<Int64>(value) == -37);
+
+  // Valid multipliers
+  ok1(InitConfigFileParser::convertStringToUint64("37k", value));
+  ok1(value == 37ull * 1024);
+  ok1(InitConfigFileParser::convertStringToUint64("37K", value));
+  ok1(value == 37ull * 1024);
+  ok1(InitConfigFileParser::convertStringToUint64("37M", value));
+  ok1(value == 37ull * 1024 * 1024);
+  ok1(InitConfigFileParser::convertStringToUint64("37G", value));
+  ok1(value == 37ull * 1024 * 1024 * 1024);
+#ifdef NOT_YET
+  ok1(InitConfigFileParser::convertStringToUint64("37T", value));
+  ok1(value == 37ull * 1024 * 1024 * 1024 * 1024);
+#endif
+
+  // Invalid multipliers
+  ok1(InitConfigFileParser::convertStringToUint64("10kB", value) == false);
+  ok1(InitConfigFileParser::convertStringToUint64("10 M", value) == false);
+  ok1(InitConfigFileParser::convertStringToUint64("10 \"M12L\"", value) == false);
+  ok1(InitConfigFileParser::convertStringToUint64("10 'M12L'", value) == false);
+  ok1(InitConfigFileParser::convertStringToUint64("10M12L", value) == false);
+  ok1(InitConfigFileParser::convertStringToUint64("10T'", value) == false);
+
+  ok1(InitConfigFileParser::convertStringToUint64("M", value) == false);
+  ok1(InitConfigFileParser::convertStringToUint64("MAX_UINT", value) == false);
+  ok1(InitConfigFileParser::convertStringToUint64("Magnus'", value) == false);
+  ok1(InitConfigFileParser::convertStringToUint64("", value) == false);
+  ok1(InitConfigFileParser::convertStringToUint64("trettisju", value) == false);
+
+  return exit_status();
+}
+
+#endif
