@@ -1,14 +1,21 @@
 /*
-   Copyright (c) 2011, 2014 Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2011, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
@@ -182,8 +189,8 @@ protected:
                            const NdbQueryOperand* const keys[],
                            const NdbQueryOptionsImpl& options,
                            const char* ident,
-                           Uint32      ix,
-                           Uint32      id,
+                           Uint32      opNo,
+                           Uint32      internalOpNo,
                            int& error);
 
   virtual const NdbQueryLookupOperationDef& getInterface() const
@@ -215,10 +222,11 @@ private:
                            const NdbQueryOperand* const keys[],
                            const NdbQueryOptionsImpl& options,
                            const char* ident,
-                           Uint32      ix,
-                           Uint32      id,
+                           Uint32      opNo,
+                           Uint32      internalOpNo,
                            int& error)
-    : NdbQueryLookupOperationDefImpl(table,keys,options,ident,ix,id,error)
+    : NdbQueryLookupOperationDefImpl(table, keys, options, ident, opNo,
+                                     internalOpNo, error)
   {}
 
   virtual NdbQueryOperationDef::Type getType() const
@@ -245,11 +253,12 @@ private:
                            const NdbQueryOperand* const keys[],
                            const NdbQueryOptionsImpl& options,
                            const char* ident,
-                           Uint32      ix,
-                           Uint32      id,
+                           Uint32      opNo,
+                           Uint32      internalOpNo,
                            int& error)
     // Add 1 to 'id' since index lookup is serialized as two operations.
-    : NdbQueryLookupOperationDefImpl(table,keys,options,ident,ix,id+1,error),
+    : NdbQueryLookupOperationDefImpl(table, keys, options, ident, opNo,
+                                     internalOpNo+1, error),
     m_index(index)
   {}
 
@@ -281,10 +290,11 @@ private:
                            const NdbTableImpl& table,
                            const NdbQueryOptionsImpl& options,
                            const char* ident,
-                           Uint32      ix,
-                           Uint32      id,
+                           Uint32      opNo,
+                           Uint32      internalOpNo,
                            int& error)
-    : NdbQueryScanOperationDefImpl(table,options,ident,ix,id,error),
+    : NdbQueryScanOperationDefImpl(table, options, ident, opNo, 
+                                   internalOpNo, error),
       m_interface(*this) 
   {}
 
@@ -385,17 +395,9 @@ NdbQueryOperandImpl& getImpl(const NdbQueryOperand& op)
 { return op.getImpl();
 }
 inline static
-NdbConstOperandImpl& getImpl(const NdbConstOperand& op)
-{ return static_cast<NdbConstOperandImpl&>(op.getImpl());
-}
-inline static
 NdbParamOperandImpl& getImpl(const NdbParamOperand& op)
 { return static_cast<NdbParamOperandImpl&>(op.getImpl());
 }
-inline static
-NdbLinkedOperandImpl& getImpl(const NdbLinkedOperand& op)
-{ return static_cast<NdbLinkedOperandImpl&>(op.getImpl());
-}
 
 /**
  * BEWARE: Return 'NULL' Until the operand has been bound to an operation.
@@ -612,24 +614,11 @@ inline static
 NdbQueryOperationDefImpl& getImpl(const NdbQueryOperationDef& op)
 { return op.getImpl();
 }
-inline static
-NdbQueryLookupOperationDefImpl& getImpl(const NdbQueryLookupOperationDef& op)
-{ return static_cast<NdbQueryLookupOperationDefImpl&>(op.getImpl());
-}
-inline static
-NdbQueryTableScanOperationDefImpl& getImpl(const NdbQueryTableScanOperationDef& op)
-{ return static_cast<NdbQueryTableScanOperationDefImpl&>(op.getImpl());
-}
-inline static
-NdbQueryIndexScanOperationDefImpl& getImpl(const NdbQueryIndexScanOperationDef& op)
-{ return static_cast<NdbQueryIndexScanOperationDefImpl&>(op.getImpl());
-}
-
 
 Uint32
-NdbQueryOperationDef::getQueryOperationIx() const
+NdbQueryOperationDef::getOpNo() const
 {
-  return ::getImpl(*this).getQueryOperationIx();
+  return ::getImpl(*this).getOpNo();
 }
 
 Uint32
@@ -702,24 +691,14 @@ NdbQueryBuilder* NdbQueryBuilder::create()
   NdbQueryBuilderImpl* const impl = new NdbQueryBuilderImpl();
   if (likely (impl != NULL))
   {
-    if (likely(impl->getNdbError().code == 0))
+    if((!ndb_join_pushdown(ndbGetOwnVersion())))
     {
-      if((!ndb_join_pushdown(ndbGetOwnVersion())))
-      {
-        /* The SPJ code is present in releases where the SPJ feature is
-         * not yet enabled.
-         */
-        impl->setErrorCode(Err_FunctionNotImplemented);
-      }
-      return &impl->m_interface;
-    }
-    else
-    {
-      // Probably failed to create Vector instances.
-      assert(impl->getNdbError().code == Err_MemoryAlloc);
-      delete impl;
-      return NULL;
+      /* The SPJ code is present in releases where the SPJ feature is
+       * not yet enabled.
+       */
+      impl->setErrorCode(Err_FunctionNotImplemented);
     }
+    return &impl->m_interface;
   }
   else
   {
@@ -892,7 +871,7 @@ NdbQueryBuilder::readTuple(const NdbDictionary::Table* table,    // Primary key
                                        options ? options->getImpl() : defaultOptions,
                                        ident,
                                        m_impl.m_operations.size(),
-                                       m_impl.getNextId(),
+                                       m_impl.getNextInternalOpNo(),
                                        error);
 
   returnErrIf(m_impl.takeOwnership(op)!=0, Err_MemoryAlloc);
@@ -966,7 +945,7 @@ NdbQueryBuilder::readTuple(const NdbDictionary::Index* index,    // Unique key l
                                        options ? options->getImpl() : defaultOptions,
                                        ident,
                                        m_impl.m_operations.size(),
-                                       m_impl.getNextId(),
+                                       m_impl.getNextInternalOpNo(),
                                        error);
 
   returnErrIf(m_impl.takeOwnership(op)!=0, Err_MemoryAlloc);
@@ -1007,7 +986,7 @@ NdbQueryBuilder::scanTable(const NdbDictionary::Table* table,
                                           options ? options->getImpl() : defaultOptions,
                                           ident,
                                           m_impl.m_operations.size(),
-                                          m_impl.getNextId(),
+                                          m_impl.getNextInternalOpNo(),
                                           error);
 
   returnErrIf(m_impl.takeOwnership(op)!=0, Err_MemoryAlloc);
@@ -1041,19 +1020,15 @@ NdbQueryBuilder::scanIndex(const NdbDictionary::Index* index,
     returnErrIf(!m_impl.m_operations[0]->isScanOperation(),
                 QRY_WRONG_OPERATION_TYPE);
 
-    // If the root is a sorted scan, we should not add another scan.
-    const NdbQueryOptions::ScanOrdering rootOrder =
-      m_impl.m_operations[0]->getOrdering();
-    returnErrIf(rootOrder == NdbQueryOptions::ScanOrdering_ascending ||
-                rootOrder == NdbQueryOptions::ScanOrdering_descending,
-                QRY_MULTIPLE_SCAN_SORTED);
-
-    // A child scan should not be sorted.
-    const NdbQueryOptions::ScanOrdering order =
-      options->getImpl().getOrdering();
-    returnErrIf(order == NdbQueryOptions::ScanOrdering_ascending ||
-                order == NdbQueryOptions::ScanOrdering_descending,
-                QRY_MULTIPLE_SCAN_SORTED);
+    if (options != NULL)
+    {
+      // A child scan should not be sorted.
+      const NdbQueryOptions::ScanOrdering order =
+        options->getImpl().getOrdering();
+      returnErrIf(order == NdbQueryOptions::ScanOrdering_ascending ||
+                  order == NdbQueryOptions::ScanOrdering_descending,
+                  QRY_MULTIPLE_SCAN_SORTED);
+    }
   }
 
   const NdbIndexImpl& indexImpl = NdbIndexImpl::getImpl(*index);
@@ -1077,7 +1052,7 @@ NdbQueryBuilder::scanIndex(const NdbDictionary::Index* index,
                                           options ? options->getImpl() : defaultOptions,
                                           ident,
                                           m_impl.m_operations.size(),
-                                          m_impl.getNextId(),
+                                          m_impl.getNextInternalOpNo(),
                                           error);
 
   returnErrIf(m_impl.takeOwnership(op)!=0, Err_MemoryAlloc);
@@ -1125,17 +1100,11 @@ NdbQueryBuilder::prepare()
 NdbQueryBuilderImpl::NdbQueryBuilderImpl()
 : m_interface(*this),
   m_error(),
-  m_operations(),
-  m_operands(),
+  m_operations(0),
+  m_operands(0),
   m_paramCnt(0),
   m_hasError(false)
-{
-  if (errno == ENOMEM)
-  {
-    // ENOMEM probably comes from Vector().
-    setErrorCode(Err_MemoryAlloc);
-  }
-}
+{}
 
 NdbQueryBuilderImpl::~NdbQueryBuilderImpl()
 {
@@ -1152,7 +1121,7 @@ NdbQueryBuilderImpl::~NdbQueryBuilderImpl()
 
 void NdbQueryBuilderImpl::setErrorCode(int aErrorCode)
 { 
-  DBUG_ASSERT(aErrorCode != 0);
+  assert(aErrorCode != 0);
   m_error.code = aErrorCode;
   m_hasError = true;
 }
@@ -1251,12 +1220,12 @@ NdbQueryDefImpl(const Vector<NdbQueryOperationDefImpl*>& operations,
                 const Vector<NdbQueryOperandImpl*>& operands,
                 int& error)
  : m_interface(*this), 
-   m_operations(operations),
-   m_operands(operands)
+   m_operations(0),
+   m_operands(0)
 {
-  if (errno == ENOMEM)
+  if (m_operations.assign(operations) || m_operands.assign(operands))
   {
-    // Failed to allocate memory for m_operations or m_operands.
+    // Failed to allocate memory in Vector::assign().
     error = Err_MemoryAlloc;
     return;
   }
@@ -1276,7 +1245,7 @@ NdbQueryDefImpl(const Vector<NdbQueryOperationDefImpl*>& operations,
   // Set length and number of nodes in tree.
   Uint32 cntLen;
   QueryTree::setCntLen(cntLen, 
-		       m_operations[m_operations.size()-1]->getQueryOperationId()+1,
+		       m_operations[m_operations.size()-1]->getInternalOpNo()+1,
 		       m_serializedDef.getSize());
   m_serializedDef.put(0,cntLen);
 
@@ -1515,7 +1484,7 @@ NdbGenericConstOperandImpl::convert2ColumnType()
   }
   else
   {
-    DBUG_ASSERT(0);
+    assert(0);
   }
 
   memcpy (dst, m_value, len);
@@ -1641,10 +1610,11 @@ NdbQueryLookupOperationDefImpl::NdbQueryLookupOperationDefImpl (
                            const NdbQueryOperand* const keys[],
                            const NdbQueryOptionsImpl& options,
                            const char* ident,
-                           Uint32      ix,
-                           Uint32      id,
+                           Uint32      opNo,
+                           Uint32      internalOpNo,
                            int& error)
-  : NdbQueryOperationDefImpl(table,options,ident,ix,id,error),
+  : NdbQueryOperationDefImpl(table, options, ident, opNo,
+                             internalOpNo, error),
    m_interface(*this)
 {
   int i;
@@ -1664,10 +1634,11 @@ NdbQueryIndexScanOperationDefImpl::NdbQueryIndexScanOperationDefImpl (
                            const NdbQueryIndexBound* bound,
                            const NdbQueryOptionsImpl& options,
                            const char* ident,
-                           Uint32      ix,
-                           Uint32      id,
+                           Uint32      opNo,
+                           Uint32      internalOpNo,
                            int& error)
-  : NdbQueryScanOperationDefImpl(table,options,ident,ix,id,error),
+  : NdbQueryScanOperationDefImpl(table, options, ident, opNo, internalOpNo, 
+                                 error),
   m_interface(*this), 
   m_index(index),
   m_paramInPruneKey(false)
@@ -1906,27 +1877,21 @@ NdbQueryOperationDefImpl::NdbQueryOperationDefImpl (
                                      const NdbTableImpl& table,
                                      const NdbQueryOptionsImpl& options,
                                      const char* ident,
-                                     Uint32      ix,
-                                     Uint32      id,
+                                     Uint32      opNo,
+                                     Uint32      internalOpNo,
                                      int& error)
   :m_isPrepared(false), 
    m_diskInChildProjection(false), 
    m_table(table), 
    m_ident(ident), 
-   m_ix(ix), m_id(id),
+   m_opNo(opNo), m_internalOpNo(internalOpNo),
    m_options(options),
    m_parent(NULL), 
-   m_children(), 
-   m_params(),
-   m_spjProjection() 
+   m_children(0), 
+   m_params(0),
+   m_spjProjection(0) 
 {
-  if (unlikely(errno == ENOMEM))
-  {
-    // Heap allocation in Vector() must have failed.
-    error = Err_MemoryAlloc;
-    return;
-  }
-  if (unlikely(m_id >= NDB_SPJ_MAX_TREE_NODES))
+  if (unlikely(m_internalOpNo >= NDB_SPJ_MAX_TREE_NODES))
   {
     error = QRY_DEFINITION_TOO_LARGE;
     return;
@@ -2132,8 +2097,8 @@ NdbQueryOperationDefImpl::appendParentList(Uint32Buffer& serializedDef) const
   if (getParentOperation() != NULL)
   {
     Uint16Sequence parentSeq(serializedDef, 1);
-    assert (getParentOperation()->getQueryOperationId() < getQueryOperationId());
-    parentSeq.append(getParentOperation()->getQueryOperationId());
+    assert (getParentOperation()->getInternalOpNo() < getInternalOpNo());
+    parentSeq.append(getParentOperation()->getInternalOpNo());
     parentSeq.finish();
     return DABits::NI_HAS_PARENT;
   }
@@ -2217,7 +2182,8 @@ NdbQueryOperationDefImpl::printTree(Uint32 depth,
   ndbout << NdbQueryOperationDef::getTypeName(getType()) << endl;
   printMargin(depth, hasMoreSiblingsMask, false);
   // Print attributes.
-  ndbout << " opNo: " << getQueryOperationIx() << endl;
+  ndbout << " opNo: " << getOpNo()
+         << " (internal: " << getInternalOpNo() << ")" << endl;
   printMargin(depth, hasMoreSiblingsMask, false);
   ndbout << " table: " << getTable().getName() << endl;
   if (getIndex() != NULL)
@@ -2256,7 +2222,7 @@ NdbQueryLookupOperationDefImpl::appendKeyPattern(Uint32Buffer& serializedDef) co
    * Key value for root operation is constructed when query is instantiated with
    * NdbQueryOperationImpl::prepareIndexKeyInfo()
    */
-  if (getQueryOperationIx() == 0)
+  if (getOpNo() == 0)
     return 0;
 
   if (m_keys[0]!=NULL)
@@ -2333,7 +2299,7 @@ NdbQueryIndexScanOperationDefImpl::appendPrunePattern(Uint32Buffer& serializedDe
    * Bound value for root operation is constructed when query is instantiated with
    * NdbQueryOperationImpl::prepareIndexKeyInfo()
    */
-  if (getQueryOperationIx() == 0)
+  if (getOpNo() == 0)
     return 0;
 
   if (m_bound.lowKeys>0 || m_bound.highKeys>0)
@@ -2517,7 +2483,7 @@ NdbQueryIndexScanOperationDefImpl::appendBoundPattern(Uint32Buffer& serializedDe
    * Bound value for root operation is constructed when query is instantiated with
    * NdbQueryOperationImpl::prepareIndexKeyInfo()
    */
-  if (getQueryOperationIx() == 0)
+  if (getOpNo() == 0)
     return 0;
 
   if (m_bound.lowKeys>0 || m_bound.highKeys>0)
@@ -2626,7 +2592,7 @@ NdbQueryPKLookupOperationDefImpl
   }
 
 #ifdef __TRACE_SERIALIZATION
-  ndbout << "Serialized node " << getQueryOperationId() << " : ";
+  ndbout << "Serialized node " << getInternalOpNo() << " : ";
   for (Uint32 i = startPos; i < serializedDef.getSize(); i++) {
     char buf[12];
     sprintf(buf, "%.8x", serializedDef.get(i));
@@ -2659,7 +2625,7 @@ NdbQueryIndexOperationDefImpl
     Uint32 requestInfo = QN_LookupNode::L_UNIQUE_INDEX;
 
     // Optional part1: Make list of parent nodes.
-    assert (getQueryOperationId() > 0);
+    assert (getInternalOpNo() > 0);
     requestInfo |= appendParentList (serializedDef);
 
     // Part2: m_keys[] are the keys to be used for index
@@ -2691,7 +2657,7 @@ NdbQueryIndexOperationDefImpl
     }
 
 #ifdef __TRACE_SERIALIZATION
-    ndbout << "Serialized index " << getQueryOperationId()-1 << " : ";
+    ndbout << "Serialized index " << getInternalOpNo()-1 << " : ";
     for (Uint32 i = startPos; i < serializedDef.getSize(); i++){
       char buf[12];
       sprintf(buf, "%.8x", serializedDef.get(i));
@@ -2717,7 +2683,7 @@ NdbQueryIndexOperationDefImpl
   // Optional part1: Append index as (single) parent op..
   { requestInfo |= DABits::NI_HAS_PARENT;
     Uint16Sequence parentSeq(serializedDef,1);
-    parentSeq.append(getQueryOperationId()-1);
+    parentSeq.append(getInternalOpNo()-1);
     parentSeq.finish();
   }
 
@@ -2747,7 +2713,7 @@ NdbQueryIndexOperationDefImpl
   }
 
 #ifdef __TRACE_SERIALIZATION
-  ndbout << "Serialized node " << getQueryOperationId() << " : ";
+  ndbout << "Serialized node " << getInternalOpNo() << " : ";
   for (Uint32 i = startPos; i < serializedDef.getSize(); i++) {
     char buf[12];
     sprintf(buf, "%.8x", serializedDef.get(i));
@@ -2764,17 +2730,17 @@ NdbQueryScanOperationDefImpl::NdbQueryScanOperationDefImpl (
                            const NdbTableImpl& table,
                            const NdbQueryOptionsImpl& options,
                            const char* ident,
-                           Uint32      ix,
-                           Uint32      id,
+                           Uint32      opNo,
+                           Uint32      internalOpNo,
                            int& error)
-  : NdbQueryOperationDefImpl(table,options,ident,ix,id,error)
+  : NdbQueryOperationDefImpl(table, options, ident, opNo, internalOpNo, error)
 {}
 
 int
 NdbQueryScanOperationDefImpl::serialize(Uint32Buffer& serializedDef,
                                         const NdbTableImpl& tableOrIndex)
 {
-  bool isRoot = (getQueryOperationIx()==0);
+  bool isRoot = (getOpNo()==0);
 
   // This method should only be invoked once.
   assert (!m_isPrepared);
@@ -2829,7 +2795,7 @@ NdbQueryScanOperationDefImpl::serialize(Uint32Buffer& serializedDef,
   }
 
 #ifdef __TRACE_SERIALIZATION
-  ndbout << "Serialized node " << getQueryOperationId() << " : ";
+  ndbout << "Serialized node " << getInternalOpNo() << " : ";
   for(Uint32 i = startPos; i < serializedDef.getSize(); i++){
     char buf[12];
     sprintf(buf, "%.8x", serializedDef.get(i));
