@@ -1,14 +1,21 @@
 /*
-   Copyright (c) 2003, 2011, Oracle and/or its affiliates. All rights reserved.
+   Copyright (c) 2003, 2021, Oracle and/or its affiliates. All rights reserved.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
@@ -22,6 +29,7 @@
 #include <NdbOut.hpp>
 
 #include <EventLogger.hpp>
+#include <NdbMutex.h>
 extern EventLogger * g_eventLogger;
 
 //#define DEBUG_OBJECTMAP
@@ -33,14 +41,24 @@ class NdbObjectIdMap
 {
 public:
   STATIC_CONST( InvalidId = 0x7fffffff );
-  NdbObjectIdMap(Uint32 initalSize, Uint32 expandSize);
+
+private:
+  /**
+   * class NdbObjectIdMap is only intended to be used through
+   * class NdbImpl.
+   */
+  friend class NdbImpl;
+
+  NdbObjectIdMap(Uint32 initialSize, Uint32 expandSize, NdbMutex* mutex);
   ~NdbObjectIdMap();
 
   Uint32 map(void * object);
   void * unmap(Uint32 id, void *object);
   
-  void * getObject(Uint32 id);
-private:
+  void * getObject(Uint32 id) const;
+
+  // mutex belonging to the NdbImpl object that owns this object map
+  NdbMutex* m_mutex;
   const Uint32 m_expandSize;
   Uint32 m_size;
   Uint32 m_firstFree;
@@ -104,8 +122,11 @@ inline
 Uint32
 NdbObjectIdMap::map(void * object)
 {
-  if(m_firstFree == InvalidId && expand(m_expandSize))
-    return InvalidId;
+  if(m_firstFree == InvalidId)
+  {
+    if (expand(m_expandSize) != 0)
+      return InvalidId;
+  }
   
   const Uint32 ff = m_firstFree;
   m_firstFree = m_map[ff].getNext();
@@ -157,11 +178,10 @@ NdbObjectIdMap::unmap(Uint32 id, void *object)
 }
 
 inline void *
-NdbObjectIdMap::getObject(Uint32 id)
+NdbObjectIdMap::getObject(Uint32 id) const
 {
   // DBUG_PRINT("info",("NdbObjectIdMap::getObject(%u) obj=0x%x", id,  m_map[id>>2].m_obj));
   id >>= 2;
-  assert(id < m_size);
   if(id < m_size)
   {
     if(m_map[id].isFree())
