@@ -1,19 +1,26 @@
-/* Copyright (c) 2003-2007 MySQL AB
-
+/*
+   Copyright (c) 2003, 2021, Oracle and/or its affiliates.
 
    This program is free software; you can redistribute it and/or modify
-   it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 of the License.
+   it under the terms of the GNU General Public License, version 2.0,
+   as published by the Free Software Foundation.
+
+   This program is also distributed with certain software (including
+   but not limited to OpenSSL) that is licensed under separate terms,
+   as designated in a particular file or component or in included license
+   documentation.  The authors of MySQL hereby grant you an additional
+   permission to link the program and your derivative works with the
+   separately licensed software that they have included with MySQL.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-   GNU General Public License for more details.
+   GNU General Public License, version 2.0, for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA */
-
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA
+*/
 
 /* ***************************************************
        FLEXHAMMER
@@ -51,7 +58,6 @@ Revision history:
 #include <ndb_global.h>
 #include <NdbApi.hpp>
 
-#include <NdbMain.h>
 #include <NdbThread.h>
 #include <NdbSleep.h>
 #include <NdbTick.h>
@@ -103,7 +109,7 @@ struct ThreadNdb {
 extern "C" void* flexHammerThread(void*);
 static int setAttrNames(void);
 static int setTableNames(void);
-static int readArguments(int, const char**);
+static int readArguments(int, char**);
 static int createTables(Ndb*);
 static int dropTables(Ndb*);
 static void sleepBeforeStartingTest(int seconds);
@@ -184,8 +190,7 @@ tellThreads(ThreadNdb* threadArrayP, const StartType what)
 
 static Ndb_cluster_connection *g_cluster_connection= 0;
  
-NDB_COMMAND(flexHammer, "flexHammer", "flexHammer", "flexHammer", 65535)
-//main(int argc, const char** argv)
+int main(int argc, char** argv)
 {
   ndb_init();
   ThreadNdb* pThreads = NULL; // Pointer to thread data array
@@ -425,7 +430,7 @@ flexHammerThread(void* pArg)
 	  tMyOpType = otDelete;
 	  break;
 	default:
-	  assert(false);
+	  require(false);
 	  break;
 	} // switch
 	    
@@ -483,7 +488,7 @@ flexHammerThread(void* pArg)
 	    check = pMyOperation[count_tables]->deleteTuple();
 	    break;
 	  default:
-	    assert(false);
+	    require(false);
 	    break;
 	  } // switch
 	  if (check == -1) {
@@ -526,7 +531,7 @@ flexHammerThread(void* pArg)
 	  case otDelete:			// Delete Case
 	    break;
 	  default:
-	    assert(false);
+	    require(false);
 	    break;
 	  } // switch
 	  if (check == -1 && tTmp == NULL && tMyOpType != otDelete) {
@@ -635,7 +640,7 @@ flexHammerThread(void* pArg)
 
 
 int
-readArguments (int argc, const char** argv)
+readArguments (int argc, char** argv)
 {
   int i = 1;
   
