@@ -24,6 +24,11 @@
 #include <NDBT_Test.hpp>
 #include <NdbRestarter.hpp>
 
+#define CHECKTRANS(trans) if ((trans) == NULL) {    \
+    ndbout << "Error at line " << __LINE__ << endl; \
+    NDB_ERR(pNdb->getNdbError());                   \
+    return NDBT_FAILED; }
+
 #define CHECKNOTNULL(p) if ((p) == NULL) {          \
     ndbout << "Error at line " << __LINE__ << endl; \
     NDB_ERR(trans->getNdbError());                  \
@@ -198,7 +203,7 @@ int testSegmentedSectionPk(NDBT_Context* ctx, NDBT_Step* step){
   NdbTransaction* trans= pNdb->startTransaction(ctx->getTab(),
                                                 &smallKey[0],
                                                 smallKeySize);
-  CHECKNOTNULL(trans);
+  CHECKTRANS(trans);
 
   /* Activate error insert 8065 in this transaction, limits
    * any single import/append to 1 section
@@ -232,9 +237,9 @@ int testSegmentedSectionPk(NDBT_Context* ctx, NDBT_Step* step){
    * on the AttrInfo section
    */
   /* Start transaction on the same node */
-  CHECKNOTNULL(trans= pNdb->startTransaction(ctx->getTab(),
-                                             &smallKey[0],
-                                             smallKeySize));
+  CHECKTRANS(trans = pNdb->startTransaction(ctx->getTab(),
+                                            &smallKey[0],
+                                            smallKeySize));
 
 
   CHECKNOTNULL(bigInsert = trans->insertTuple(record, bigAttrRowBuf));
@@ -251,9 +256,9 @@ int testSegmentedSectionPk(NDBT_Context* ctx, NDBT_Step* step){
    * in the same batch.  Check that abort handling is correct
    */
     /* Start transaction on the same node */
-  CHECKNOTNULL(trans= pNdb->startTransaction(ctx->getTab(),
-                                             &smallKey[0],
-                                             smallKeySize));
+  CHECKTRANS(trans = pNdb->startTransaction(ctx->getTab(),
+                                            &smallKey[0],
+                                            smallKeySize));
   /* First op in batch, will cause overflow */
   CHECKNOTNULL(bigInsert = trans->insertTuple(record, bigAttrRowBuf));
   
@@ -273,9 +278,9 @@ int testSegmentedSectionPk(NDBT_Context* ctx, NDBT_Step* step){
    * with a big key value
    */
   /* Start transaction on the same node */
-  CHECKNOTNULL(trans= pNdb->startTransaction(ctx->getTab(),
-                                             &smallKey[0],
-                                             smallKeySize));
+  CHECKTRANS(trans = pNdb->startTransaction(ctx->getTab(),
+                                            &smallKey[0],
+                                            smallKeySize));
   
   NdbOperation* bigInsertOldApi;
   CHECKNOTNULL(bigInsertOldApi= trans->getNdbOperation(ctx->getTab()));
@@ -302,9 +307,9 @@ int testSegmentedSectionPk(NDBT_Context* ctx, NDBT_Step* step){
   /* Now try with a 'short' TCKEYREQ, generated using the old Api 
    * with a big data value
    */
-  CHECKNOTNULL(trans= pNdb->startTransaction(ctx->getTab(),
-                                             &smallKey[0],
-                                             smallKeySize));
+  CHECKTRANS(trans = pNdb->startTransaction(ctx->getTab(),
+                                            &smallKey[0],
+                                            smallKeySize));
   
   CHECKNOTNULL(bigInsertOldApi= trans->getNdbOperation(ctx->getTab()));
 
@@ -348,9 +353,9 @@ int testSegmentedSectionPk(NDBT_Context* ctx, NDBT_Step* step){
    * short key, but not the AttrInfo
    */
   /* Start transaction on same node */
-  CHECKNOTNULL(trans= pNdb->startTransaction(ctx->getTab(),
-                                             &smallKey[0],
-                                             smallKeySize));
+  CHECKTRANS(trans = pNdb->startTransaction(ctx->getTab(),
+                                            &smallKey[0],
+                                            smallKeySize));
   
   CHECKNOTNULL(bigInsertOldApi= trans->getNdbOperation(ctx->getTab()));
   
@@ -379,9 +384,9 @@ int testSegmentedSectionPk(NDBT_Context* ctx, NDBT_Step* step){
   restarter.insertErrorInAllNodes(8067);
 
   /* Now a 'short' TCKEYREQ - there will be no space to import the key */
-  CHECKNOTNULL(trans= pNdb->startTransaction(ctx->getTab(),
-                                             &smallKey[0],
-                                             smallKeySize));
+  CHECKTRANS(trans = pNdb->startTransaction(ctx->getTab(),
+                                            &smallKey[0],
+                                            smallKeySize));
   
   CHECKNOTNULL(bigInsertOldApi= trans->getNdbOperation(ctx->getTab()));
   
@@ -408,9 +413,9 @@ int testSegmentedSectionPk(NDBT_Context* ctx, NDBT_Step* step){
   /* Finished with error insert, cleanup the error insertion
    * Error insert 8068 will free the hoarded segments
    */
-  CHECKNOTNULL(trans= pNdb->startTransaction(ctx->getTab(),
-                                             &smallKey[0],
-                                             smallKeySize));
+  CHECKTRANS(trans = pNdb->startTransaction(ctx->getTab(),
+                                            &smallKey[0],
+                                            smallKeySize));
 
   CHECKEQUAL(NDBT_OK, activateErrorInsert(trans, 
                                           record, 
@@ -570,7 +575,7 @@ int testSegmentedSectionIx(NDBT_Context* ctx, NDBT_Step* step){
   trans= pNdb->startTransaction(ctx->getTab(),
                                 &smallKey[0],
                                 smallKeySize);
-  CHECKNOTNULL(trans);
+  CHECKTRANS(trans);
 
   /* Activate error insert 8065 in this transaction, limits any
    * single append/import to 10 sections.
@@ -606,9 +611,9 @@ int testSegmentedSectionIx(NDBT_Context* ctx, NDBT_Step* step){
    * operations exist in this new transaction.
    */
   /* Start a transaction on a specific node */
-  CHECKNOTNULL(trans= pNdb->startTransaction(ctx->getTab(),
-                                             &smallKey[0],
-                                             smallKeySize));
+  CHECKTRANS(trans = pNdb->startTransaction(ctx->getTab(),
+                                            &smallKey[0],
+                                            smallKeySize));
 
   CHECKNOTNULL(trans->readTuple(ixRecord,
                                 bigKeyIxBuf,
@@ -625,9 +630,9 @@ int testSegmentedSectionIx(NDBT_Context* ctx, NDBT_Step* step){
   /* Now a TCINDXREQ that overflows, but is not the last in the
    * batch, what happens to the other TCINDXREQ in the batch?
    */
-  CHECKNOTNULL(trans= pNdb->startTransaction(ctx->getTab(),
-                                             &smallKey[0],
-                                             smallKeySize));
+  CHECKTRANS(trans = pNdb->startTransaction(ctx->getTab(),
+                                            &smallKey[0],
+                                            smallKeySize));
 
   CHECKNOTNULL(trans->readTuple(ixRecord,
                                 bigKeyIxBuf,
@@ -653,9 +658,9 @@ int testSegmentedSectionIx(NDBT_Context* ctx, NDBT_Step* step){
    * handling TRANSID_AI
    */
   /* Start a transaction on a specific node */
-  CHECKNOTNULL(trans= pNdb->startTransaction(ctx->getTab(),
-                                             &smallKey[0],
-                                             smallKeySize));
+  CHECKTRANS(trans = pNdb->startTransaction(ctx->getTab(),
+                                            &smallKey[0],
+                                            smallKeySize));
 
   /* Activate error insert 8066 in this transaction, limits a
    * single import/append to 1 section.
@@ -696,9 +701,9 @@ int testSegmentedSectionIx(NDBT_Context* ctx, NDBT_Step* step){
   /* Now try with a 'short' TCINDXREQ, generated using the old Api 
    * with a big index key value
    */
-  CHECKNOTNULL(trans= pNdb->startTransaction(ctx->getTab(),
-                                             &smallKey[0],
-                                             smallKeySize));
+  CHECKTRANS(trans = pNdb->startTransaction(ctx->getTab(),
+                                            &smallKey[0],
+                                            smallKeySize));
   
   const NdbDictionary::Index* index;
   CHECKNOTNULL(index= pNdb->getDictionary()->
@@ -728,9 +733,9 @@ int testSegmentedSectionIx(NDBT_Context* ctx, NDBT_Step* step){
   /* Now try with a 'short' TCINDXREQ, generated using the old Api 
    * with a big attrinfo value
    */
-  CHECKNOTNULL(trans= pNdb->startTransaction(ctx->getTab(),
-                                             &smallKey[0],
-                                             smallKeySize));
+  CHECKTRANS(trans = pNdb->startTransaction(ctx->getTab(),
+                                            &smallKey[0],
+                                            smallKeySize));
   
   NdbIndexOperation* bigUpdateOldApi;
   CHECKNOTNULL(bigUpdateOldApi= trans->getNdbIndexOperation(index));
@@ -764,9 +769,9 @@ int testSegmentedSectionIx(NDBT_Context* ctx, NDBT_Step* step){
   /* Now a short TCINDXREQ where the KeyInfo from the TCINDXREQ
    * can be imported, but the ATTRINFO can't
    */
-  CHECKNOTNULL(trans= pNdb->startTransaction(ctx->getTab(),
-                                             &smallKey[0],
-                                             smallKeySize));
+  CHECKTRANS(trans = pNdb->startTransaction(ctx->getTab(),
+                                            &smallKey[0],
+                                            smallKeySize));
   
   CHECKNOTNULL(bigUpdateOldApi= trans->getNdbIndexOperation(index));
 
@@ -797,9 +802,9 @@ int testSegmentedSectionIx(NDBT_Context* ctx, NDBT_Step* step){
   /* Now a short TCINDXREQ where the KeyInfo from the TCINDXREQ
    * can't be imported
    */
-  CHECKNOTNULL(trans= pNdb->startTransaction(ctx->getTab(),
-                                             &smallKey[0],
-                                             smallKeySize));
+  CHECKTRANS(trans = pNdb->startTransaction(ctx->getTab(),
+                                            &smallKey[0],
+                                            smallKeySize));
 
   CHECKNOTNULL(bigUpdateOldApi= trans->getNdbIndexOperation(index));
 
@@ -827,9 +832,9 @@ int testSegmentedSectionIx(NDBT_Context* ctx, NDBT_Step* step){
 #endif  
 
   /* Finished with error insert, cleanup the error insertion */
-  CHECKNOTNULL(trans= pNdb->startTransaction(ctx->getTab(),
-                                             &smallKey[0],
-                                             smallKeySize));
+  CHECKTRANS(trans = pNdb->startTransaction(ctx->getTab(),
+                                            &smallKey[0],
+                                            smallKeySize));
 
   CHECKEQUAL(NDBT_OK, activateErrorInsert(trans, 
                                           baseRecord, 
@@ -892,7 +897,7 @@ int testSegmentedSectionScan(NDBT_Context* ctx, NDBT_Step* step){
   NdbTransaction* trans= pNdb->startTransaction(ctx->getTab(),
                                                 &smallKey[0],
                                                 smallKeySize);
-  CHECKNOTNULL(trans);
+  CHECKTRANS(trans);
 
   /* Activate error insert 8066 in this transaction, limits a 
    * single import/append to 1 section.
@@ -938,9 +943,9 @@ int testSegmentedSectionScan(NDBT_Context* ctx, NDBT_Step* step){
   trans->close();
 
   /* Finished with error insert, cleanup the error insertion */
-  CHECKNOTNULL(trans= pNdb->startTransaction(ctx->getTab(),
-                                             &smallKey[0],
-                                             smallKeySize));
+  CHECKTRANS(trans = pNdb->startTransaction(ctx->getTab(),
+                                            &smallKey[0],
+                                            smallKeySize));
 
   CHECKEQUAL(NDBT_OK, activateErrorInsert(trans, 
                                           record, 
@@ -1001,7 +1006,7 @@ int testDropSignalFragments(NDBT_Context* ctx, NDBT_Step* step){
   {
     /* Start a transaction */
     NdbTransaction* trans= pNdb->startTransaction();
-    CHECKNOTNULL(trans);
+    CHECKTRANS(trans);
 
     SubCase subcase= cases[iteration % numSubCases];
 
