@@ -256,11 +256,6 @@ extra_args=""
 if test -r "$basedir/my.cnf"
 then
   extra_args="-e $basedir/my.cnf"
-else
-  if test -r "$datadir/my.cnf"
-  then
-    extra_args="-e $datadir/my.cnf"
-  fi
 fi
 
 parse_server_arguments `$print_defaults $extra_args mysqld server mysql_server mysql.server`
@@ -311,6 +306,9 @@ case "$mode" in
 
     if test -s "$mysqld_pid_file_path"
     then
+      # signal mysqld_safe that it needs to stop
+      touch "$mysqld_pid_file_path.shutdown"
+
       mysqld_pid=`cat "$mysqld_pid_file_path"`
 
       if (kill -0 $mysqld_pid 2>/dev/null)
